#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Examples of scopes and tests to show results of scope changes
    between global, local, and nonlocal scopes. """


def scope_test():
    def do_local():
        spam = "local spam (l in l)"
        print("In local env: ", spam)
        print("local var is printed")
        print(f"{id(spam)=}")

    def do_nonlocal():
        nonlocal spam
        spam = "nonlocal spam (n in l)"
        print("In local env: ", spam)
        print(f"{id(spam)=}")

    def do_global():
        global spam
        spam = "global spam (g in l)"

    spam = "test spam(l in l ... in scope_test)"
    print("Before: ", spam)
    print()

    # spam assigned to "local spam" in local env
    do_local()
    print("After local assignment:", spam)
    print()

    # spam assigned to "nonlocal spam" as nonlocal var
    do_nonlocal()
    print("After nonlocal assignment:", spam)
    print()

    # spam assigned to "local spam" in local env
    do_global()
    print("After global assignment:", spam)
    print()


# ? ------------------------ tests

spam: str = "initial spam (g in g)"
print("Initial spam (global in global env): ", spam)
scope_test()
print("In global scope:", spam)
print(
    """The global variable was carried into the 'scope_test'
      function, changed to (l in l), """
)



def get_encoding():
    from locale import getpreferredencoding

    try:
        return getpreferredencoding(do_setlocale=True) or "utf-8"
    except:
        return "utf-8"


DEFAULT_ENCODING: str = get_encoding()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 'future imports'
from __future__ import absolute_import, print_function

# 'Standard Library'
import codecs
import os
import re

# 'third party'
from setuptools import find_packages, setup

#
# Copyright 2007-2016 The Python-Twitter Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.




cwd = os.path.abspath(os.path.dirname(__file__))


def read(filename):
    with codecs.open(os.path.join(cwd, filename), "rb", "utf-8") as h:
        return h.read()


metadata = read(os.path.join(cwd, "twitter", "__init__.py"))


def extract_metaitem(meta):
    # swiped from https://hynek.me 's attr package
    meta_match = re.search(
        r"""^__{meta}__\s+=\s+['\"]([^'\"]*)['\"]""".format(meta=meta),
        metadata,
        re.MULTILINE,
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


setup(
    name="python-twitter",
    version=extract_metaitem("version"),
    license=extract_metaitem("license"),
    description=extract_metaitem("description"),
    long_description=(
        read("README.rst")
        + "\n\n"
        + read("AUTHORS.rst")
        + "\n\n"
        + read("CHANGES")
    ),
    author=extract_metaitem("author"),
    author_email=extract_metaitem("email"),
    maintainer=extract_metaitem("author"),
    maintainer_email=extract_metaitem("email"),
    url=extract_metaitem("url"),
    download_url=extract_metaitem("download_url"),
    packages=find_packages(exclude=("tests", "docs")),
    platforms=["Any"],
    install_requires=["requests", "requests-oauthlib"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    keywords="twitter api",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# !---------------------------------------------- Custom Types


class Const:
    """ A wrapper that makes variables constant. """

    def __init__(self, val, report_errors=True):
        super(Const, self).__setattr__("value", val)
        self.report_errors: bool = report_errors

    def __setattr__(self, name, val):
        if self.report_errors:
            raise ValueError("Trying to change a constant value!", self)

    def __str__(self):
        return str(self.value)


class _Struct(type):
    """ A custom structure that can have any fields defined. """

    # Ref: https://stackoverflow.com/questions/35988/c-like-structures-in-python

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class _MapDict(dict):
    """ A custom dictionary that autoloads it's kwargs. """

    def __init__(self, **kwargs):
        super(_MapDict, self).__init__(**kwargs)
        self.__dict__ = self


class _KwargStruct(type):
    """ A custom type that creates instance variables from args and kwargs.

        e.g.
        ```
        class MyClass(object):
        ''' An example custom class based on _KwargStruct. '''
        __metaclass__ = _KwargStruct

        def __init__(self, a, b, c):
            pass
        ```
        """

    def __call__(cls, *args, **kwargs):
        names = cls.func_code.co_varnames[1:]

        self = type.__call__(cls, *args, **kwargs)

        for name, value in zip(names, args):
            setattr(self, name, value)

        for name, value in kwargs.items():
            setattr(self, name, value)
        return self


def _init_all_args(fn):  # TODO --- not working (renamed as _xxx)
    """ A custom decorator used to wrap __init__ in custom classes in
        order to create instance variables from args and kwargs.

        e.g.
        ```
        class Test(object):
        @init_all_args
        def __init__(self, a, b):
            pass
        ```
        """

    @wraps(fn)
    def wrapped_init(self, *args, **kwargs):
        """ TODO --- AttributeError: 'function' object has no attribute 'func_code' """

        names = fn.func_code.co_varnames[1:]
        print(f"{names=}")
        vprint("names")

        for name, value in zip(names, args):
            setattr(self, name, value)

        for name, value in kwargs.items():
            setattr(self, name, value)

    return wrapped_init


# Singleton/BorgSingleton.py
# Alex Martelli's 'Borg'


class Borg:
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class Singleton(Borg):
    def __init__(self, arg):
        Borg.__init__(self)
        self.val = arg

    def __str__(self):
        return self.val


stuff = "the stuff"
other_stuff = "the other stuff"
vprint("stuff", "other_stuff")



# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
__license__ = "MIT"


class WebPageError(TypeError):
    """  # Exception raised for errors in the WebPage class. """

    pass


class WebPage(requests.Response):
    """ A local representation of a webpage. This object is able to analyze itself, share data, and perform its duties unsupervised.

    parameters

    url: the source of the original page grab. Use the <refresh> command to get an updated version."""

    # >>> r = requests.get('https://api.github.com/user',

    def __init__(self, url):
        super().__init__()
        self.url: str = url
        self.dirty: bool = True
        self.last_status: int = 0
        self.timestamp: struct_time = gmtime()
        # self.tags = {}
        # self.auth: Tuple[str, str] = ('user', 'pass')
        # auth = ('user', 'pass')

    def get_url_content(self, url: str) -> str:
        """ Return decoded contents from <url> using default parameters. """
        self = requests.get(url)
        self.history.append()
        self.last_status = self.status_code
        if self.last_status == 200:
            return self.text
        else:
            return ""

    def tag_find(
        self,
        tag_name: str,
        attrs_pass: Dict[Any, Any],
        parser_pass: str = DEFAULT_PARSER,
    ) -> List[Any]:
        """ Find matching tags from url. """
        soup = BeautifulSoup(self.text, features=DEFAULT_PARSER)
        return soup.findAll(name=tag_name, attrs=attrs_pass)

    def tag_list(self, tag_name: str):
        if self.dirty:
            self.tags = Counter(self.text)
            return self.tags

    def to_markdown(self):
        pass

    def to_json(self):
        return json.dumps(self.text)

    def stats(self):
        pass

    def soup(self):
        pass  # return BeautifulSoup()


class WebPageSet(deque):
    """ Stores and maintains a set of web pages. """

    DEFAULT_WEBPAGESET_SIZE = 2000  # maximum number of pages
    # TODO this should be 'maximum size' ... not count
    # ... and class should check it's own size

    def __init__(
        self,
        iterable: MutableSequence,
        maxlen: int = 0,
        check_links: bool = True,
        image_storage: str = "",
    ):
        if not maxlen or maxlen < 1:
            maxlen = DEFAULT_WEBPAGESET_SIZE
        self.check_links = check_links
        self.store = True if image_storage else False
        self.image_storage = image_storage
        super().__init__(iterable, maxlen)

    def count(self):
        """ count tags, emails, ... whatever from the entire set """
        pass

    def common(self):
        """ find items that this pageset has in common. """
        pass

    def append(self, x):
        if isinstance(x, WebPage):
            super().append(x)
        else:
            if SET_DEBUG:
                db_print()
            else:
                raise (WebPageError)

    def size_check(self):
        print(self.__sizeof__())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" as_constants.py

    copyright (c) 2019 Michael Treanor
    https://www.github.com/skeptycal
    https://www.twitter.com/skeptycal
    """

if True:
    import locale
    import math
    import os
    import pathlib
    import platform
    import sys
    from typing import Any, Dict, FrozenSet, List, Sequence, Tuple

# * @AUTOSYS_PARSE: START

locale.setlocale(locale.LC_ALL, "")
if True:
    ##############################################
    # custom types
    borderType: FrozenSet[str] = {"single", "double", "graphic", "text"}

    ##############################################
    # general constants
    SET_DEBUG: bool = True
    PI: float = math.pi

    ##############################################
    # working variables
    i: int = 0
    n: int = 0
    filename: str = ""
    file_list: List[str] = []
    word_list: List[str] = []

    ##############################################
    # Python specific constants
    _PY2: bool = sys.version_info[0] == 2
    PY_ENV: os._Environ = os.environ
    PY_BASE: str = os.path.basename(PY_ENV["_"])
    # PY_VER: str = ".".join(str(_) for _ in __import__("sys").version_info[:3])
    PY_VER: str = sys.version.split(" ")[0]
    _PY3: bool = (lambda x: [x for x in [False]] and None or x)(True)
    PYTHON_PATH: List[str] = PY_ENV["PYTHONPATH"].split(os.pathsep)

    ##############################################
    # System Constants
    SEP: str = os.sep
    LINESEP: str = os.linesep
    LF: str = chr(10)
    IS_64BITS = sys.maxsize > 2 ** 32
    PWD = pathlib.Path.cwd().resolve()

    from unicodedata import name as UCNAME

    ASCII_CHARS = {
        "-",
        ")",
        "%",
        "!",
        "=",
        "9",
        "5",
        "1",
        "M",
        "I",
        "E",
        "A",
        "]",
        "Y",
        "U",
        "Q",
        "m",
        "i",
        "e",
        "a",
        "}",
        "y",
        "u",
        "q",
        ",",
        "(",
        "$",
        " ",
        "<",
        "8",
        "4",
        "0",
        "L",
        "H",
        "D",
        "@",
        "\\",
        "X",
        "T",
        "P",
        "l",
        "h",
        "d",
        "`",
        "|",
        "x",
        "t",
        "p",
        "/",
        "+",
        "'",
        "#",
        "?",
        ";",
        "7",
        "3",
        "O",
        "K",
        "G",
        "C",
        "_",
        "[",
        "W",
        "S",
        "o",
        "k",
        "g",
        "c",
        "{",
        "w",
        "s",
        ".",
        "*",
        "&",
        '"',
        ">",
        ":",
        "6",
        "2",
        "N",
        "J",
        "F",
        "B",
        "^",
        "Z",
        "V",
        "R",
        "n",
        "j",
        "f",
        "b",
        "~",
        "z",
        "v",
        "r",
    }


class DEFAULT:
    """ Autosys package defaults """

    # Environment constants
    DICT_DISPLAY_SEPARATOR: str = ": "
    ENCODING: str = locale.getlocale()[1] if locale.getlocale()[1] else "UTF-8"
    LANG: str = locale.getlocale()[0] if locale.getlocale()[0] else "en_US"
    PREFERRED_ENCODING: str = locale.getpreferredencoding()

    CONNECT_ERRORS: Tuple[str] = (
        ConnectionError,
        ConnectionAbortedError,
        ConnectionResetError,
        ConnectionRefusedError,
    )


class CLI:
    DISPLAY_WIDTH: int = 80
    FIELD_PADDING: int = 15


class AS_MATH:
    """ Math shortcuts and efficient functions """

    # example lookup tables
    SIN: Dict[float, float] = {
        a / 10: math.sin(math.radians(a / 10)) for a in range(3600)
    }
    COS: Dict[float, float] = {
        a / 10: math.cos(math.radians(a / 10)) for a in range(3600)
    }
    TAN: Dict[float, float] = {
        a / 10: math.tan(math.radians(a / 10)) for a in range(3600)
    }


class C_ERRORS:
    """
    C++ style error messages
        Reference:
        Advanced Bash-Scripting Guide
        <http://tldp.org/LDP/abs/html/exitcodes.html#EXITCODESREF>
        from /usr/include/sysexits.h
        Copyright (c) 1987, 1993
        The Regents of the University of California.  All rights reserved.
    """

    EX_OK = (0,)  # successful termination
    EX__BASE = (64,)  # base value for error messages
    EX_USAGE = (64,)  # command line usage error
    EX_DATAERR = (65,)  # data format error
    EX_NOINPUT = (66,)  # cannot open input
    EX_NOUSER = (67,)  # addressee unknown
    EX_NOHOST = (68,)  # host name unknown
    EX_UNAVAILABL = (69,)  # service unavailable
    EX_SOFTWARE = (70,)  # internal software error
    EX_OSERR = (71,)  # system error (e.g., can't fork)
    EX_OSFILE = (72,)  # critical OS file missing
    EX_CANTCREAT = (73,)  # can't create (user) output file
    EX_IOERR = (74,)  # input/output error
    EX_TEMPFAIL = (75,)  # temp failure; user is invited to retry
    EX_PROTOCOL = (76,)  # remote error in protocol
    EX_NOPERM = (77,)  # permission denied
    EX_CONFIG = (78,)  # configuration error
    EX__MAX = 78  # maximum listed value


__all__ = ["Default", "CLI", "AS_MATH", "C_ERRORS"]

# * @AUTOSYS_PARSE: STOP
if __name__ == "__main__":
    print()
    print(PY_ENV.get("CANARY"))
    print("Test Values")
    print("*" * 40)
    print(PY_ENV.get("RESET_FG"))

    print("name: ", name)
    print("version: ", version)
    print("DEFAULT.ENCODING: ", DEFAULT.ENCODING)
    print("python version 3+: ", PY3)
    print("PY_VER: ", PY_VER)
    print("SIN(0): ", AS_MATH.SIN.get(math.radians(0)))
    print("PYTHON_PATH[0]: ", PYTHON_PATH[0])
    print("PI: ", PI)
    print()
    print("__doc__: ", __doc__)
    print("__file__: ", __file__)
    print("__package__: ", __package__)
    print("__name__: ", __name__)
    print("PY_VER: ", PY_VER)

    p = PPath(PWD)
    print(p.str())
    print(p.next_line())



def dunder_it(string: str,
              fill_spaces: bool = True,
              ignore_errors: bool = True) -> (str, Exception):
    ''' Return name `string` with dunder prefix and suffix added if needed.

        fill_spaces - if True, replace spaces with '_'

        ignore_errors - if True, return empty string on errors

        e.g.
        ```
        NL -> __NL__
        the_variable -> __the_variable__
        __doc__ -> __doc__
        ```
        '''
    try:
        return f"__{str(string).replace(' ','_').lstrip(' _').rstrip(' _')}__"
    except Exception as e:
        if ignore_errors:
            return ""  # skip errors
        return e  # return errors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" as_80col.py """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal


# 'Standard Library'
import fileinput
import pathlib
import sys
import textwrap

# 'package imports'
from autosys.as_constants import DEFAULT

from typing import List


def create_test_file(
    filename: str = "testfile.txt",
    encoding: str = as_constants.DEFAULT_ENCODING,
) -> str:
    """ Create test file and return resolved Purepath """
    p = pathlib.Path(filename)
    p = p.resolve()
    with open(p, mode="w", encoding=encoding, errors="ignore") as f:
        for n in range(20):
            line_str = "X" * (n + 70) + "\n"
            # print(line_str)
            f.write(line_str)
    return p


def trunc_string_width(s: str) -> str:
    return s


if __name__ == "__main__":

    test_value = """This function wraps the input paragraph such that each line
    in the paragraph is at most width characters long. The wrap method
    returns a list of output lines. The returned list
    is empty if the wrapped
    output has no content."""

    set_width = DEFAULT_CLI_WIDTH
    print("Default Language: ", DEFAULT_LANG)
    print("Default Encoding: ", DEFAULT_ENCODING)
    exit

    if len(sys.argv) < 2:
        print("No arguments given. Creating testfile.txt.")
        file_list.append(create_test_file("testfile.txt"))
        show_test = 1
    else:
        try:
            arg_int = int(sys.argv[1])
        except (IndexError, ValueError):
            # print("Unable to parse code as an integer")
            file_list = sys.argv[1:]
        else:
            # print("first argument is int")
            set_width = arg_int
            file_list = sys.argv[2:]

    # with fileinput.input(file_list, inplace=True) as f:
    #     for line in f:
    #         # process(line)

    #         value: str = line.strip()
    #         # Wrap this text.
    #         wrapper = textwrap.TextWrapper(width=79)
    #         word_list = "\n".join(wrapper.wrap(value))

    #         # Process the line here
    #         print(word_list)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Michael Treanor
# License: MIT (http://www.opensource.org/licenses/mit-license.php)
"""
testing some things with strings and fstrings ... with wings

https://www.python.org/dev/peps/pep-0498/

fstring definition:
    ```
    f ' <text> { <expression> <optional !s, !r, or !a> <optional : format specifier> } <text> ... '
    ```
"""


class NewStr(str):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" dis_block.py """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal


if True:  # imports
    import dis
    import sys
    from pathlib import Path
    from collections import Counter, deque
    from time import perf_counter_ns as timer
    from typing import Deque, Dict, List

    try:
        import ujson as json
    except ImportError:
        import json
if True:  # constants

    _PY2: bool = sys.version_info[0] == 2

    class CB_Options:
        """ Initialization option choices for CodeBlock objects. """

        ON_DEMAND = 0  # minimal - only processed as needed (jit)
        ON_INIT = 1  # basic - only generated on object creation (precompiled)
        ON_ALL = 2  # thorough - processed when created and refreshed on demand

    CB_OPTIONS_DEFAULT: CB_Options = CB_Options.ON_DEMAND


class CodeBlock:
    _code: str
    _dis: str
    _name: str
    _result: str

    def __init__(self, name: str, s: str = "", options=CB_Options.ON_DEMAND):
        self.name = name
        self.code = self.name if s == "" else s

    @property
    def dis(self):
        return dis.dis(self.code)

    @property
    def result(self):
        return eval(self.code)

    @property
    def compile(self):
        return compile(self.code)

    def __str__(self):
        return self.name


class TestRun:
    trials: int = 1
    repeats: int = 100
    avgtime: Deque


class CodeBlocks:
    """ Collection of CodeBlock objects with test run instructions.
        Used to load and setup CodeBlock objects before profiling runs.
        """

    __data = {}
    __cb_list: List = []
    _file = ""
    _options: CB_Options

    def __init__(self, file=None, opt: CB_Options = CB_OPTIONS_DEFAULT):
        super().__init__()
        self._options = opt
        if Path(file).exists():
            self._file = file

    @property
    def options(self):
        return self._options

    def load_dict(self, d: Dict):
        """ Load data from dict. """
        for n, k, v in enumerate(d):
            self.cb_list.append(CodeBlock(k, v, self.options))
            pass
        self.data = d

    def save_json(self, f):
        """ Save data to json file. """
        pass

    def print(self):
        for cb in cb_list:
            print(cb)


code_tests = {"_PY2": "", "sys.argv": ""}

# print(A)
# print(A.result)

# print(f"disassemble test A: {dis_test_A}")
# print(f"result of A = {dis_test_A()}")
# print(dis.dis(dis_test_A))

# print(f"disassemble test B: {dis_test_B}")
# print(f"result of B = {dis_test_B()}")
# print(dis.dis(dis_test_B))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# TODO - this is just a rough idea ... not yet implemented ... not functional

# 'Standard Library'
import sys

from io import TextIOWrapper

# 'package imports'
from cli.anansi import *
from dev.debug import log, logit
from profiling.timeit import *

from typing import List

LOG_FILE: TextIOWrapper = sys.stdout
__all__: List[str] = []
ansi = Ansi


def a(func):
    def _all(*args, **kwargs):
        result = func(*args, **kwargs)
        name = func.__name__
        print(f"a({name}) is {result}")
        if name not in __all__:
            __all__.append(name)
        return result

    return _all


@timeit
@try_it
@a
def test_all_decorator(*args, **kwargs):
    print(ansi.BBRIGHTBLUE, " ** test all the decorators")
    print(1 / 0)


test_all_decorator("try this stuff")
print(f"__all__ = {__all__}")



import contextlib
import functools
import os
import shutil
import tempfile

from pathlib import Path

from typing import AnyStr


@contextlib.contextmanager
def temp_dir(
    suffix: AnyStr = "",
    prefix: AnyStr = "tmp",
    parent_dir: AnyStr = None,
    make_cwd=False,
):
    """ Create a temporary directory and optionally change the current
        working directory to it. The directory is deleted when the context
        exits.

        The temporary directory is created when entering the context
        manager, and deleted when exiting it:
        >>> import temporary
        >>> with temporary.temp_dir() as temp_dir:
        ...     assert temp_dir.is_dir()
        >>> assert not temp_dir.exists()

        This time let's make the temporary directory our working directory:
        >>> import os
        >>> with temporary.temp_dir(make_cwd=True) as temp_dir:
        ...     assert str(temp_dir) == os.getcwd()
        >>> assert not str(temp_dir) == os.getcwd()

        The suffix, prefix, and parent_dir options are passed to the
        standard ``tempfile.mkdtemp()`` function:
        >>> with temporary.temp_dir() as p:
        ...     with temporary.temp_dir(suffix='suf', prefix='pre'
        parent_dir=p) as d:
        ...         assert d.parent == p
        ...         assert d.name.startswith('pre')
        ...         assert d.name.endswith('suf')

        This function can also be used as a decorator, with the in_temp_dir
        alias:
        >>> @temporary.in_temp_dir()
        ... def my_function():
        ...     assert old_cwd != os.getcwd()
        ...
        >>> old_cwd = os.getcwd()
        >>> my_function()
        >>> assert old_cwd == os.getcwd()
        """
    prev_cwd = Path().cwd()

    # if not parent directory is given, use cwd
    parent_dir = Path(parent_dir).resolve() if parent_dir else prev_cwd

    tmp_dir = tempfile.TemporaryDirectory(suffix=suffix, prefix=prefix, dir=parent_dir)

    abs_path = tempfile.mkdtemp(suffix, prefix, parent_dir)
    path = Path(abs_path)
    try:
        if make_cwd:
            os.chdir(str(abs_path))
        yield path.resolve()
    finally:
        if make_cwd:
            os.chdir(prev_cwd)
        with temporary.util.allow_missing_file():
            shutil.rmtree(str(abs_path))


in_temp_dir = functools.partial(temp_dir, make_cwd=True)  # pylint: disable=invalid-name



#!/usr/bin/env python3
'''
Small wsgiref based web server. Takes a path to serve from and an
optional port number (defaults to 8000), then tries to serve files.
Mime types are guessed from the file names, 404 errors are raised
if the file is not found. Used for the make serve target in Doc.
'''
import sys
import os
import mimetypes
from os import linesep as NL
from sys import stderr, stdout, path as PYTHONPATH
from pathlib import Path

here = Path(__file__).resolve().parent
parents = Path(__file__).resolve().parents

for parent in parents:
    if Path(parent/'README.md').exists():
        break
    if parent not in PYTHONPATH:
        print(parent)
        PYTHONPATH.insert(0, parent)

if here not in PYTHONPATH:
    PYTHONPATH.insert(0, here)

while True:
    try:
        continue
    except KeyboardInterrupt:
        break

if True:
    from wsgiref import simple_server, util

    import autosys.web.webpages.make_html_index

try:
    DEFAULT_ENCODING
except:
    from locale import getpreferredencoding
    DEFAULT_ENCODING: str = getpreferredencoding(do_setlocale=True) or "utf-8"


def app(environ, respond):
    fn = os.path.join(path, environ['PATH_INFO'][1:])
    if '.' not in fn.split(os.path.sep)[-1]:
        fn = os.path.join(fn, 'index.html')
    type = mimetypes.guess_type(fn)[0]

    if os.path.exists(fn):
        respond('200 OK', [('Content-Type', type)])
        return util.FileWrapper(open(fn, "rb"))
    else:
        respond('404 Not Found', [('Content-Type', 'text/plain')])
        return [b'not found']


if __name__ == '__main__':
    makehtmlindex()
    # print(f"{get_html_file()}")
    pass
if True:
    path = sys.argv[1] if len(sys.argv) > 1 else os.getcwd()
    port = int(sys.argv[2]) if len(sys.argv) > 2 else 8000
    httpd = simple_server.make_server('', port, app)
    print("Serving {} on port {}, control-C to stop".format(path, port))
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("Shutting down.")
        httpd.server_close()



# 'Standard Library'
import logging
import os

from dataclasses import dataclass
from pathlib import Path, _posix_flavour, _windows_flavour

# 'package imports'
from autosys.implore.text import TextMessaging


@dataclass(frozen=True)
class Verbosity:
    """ Reports output of the selected level and below. """

    V_CRITICAL: int = 50  # Serious error - unable to continue
    V_ERROR: int = 40  # Failure of essential function
    V_WARN: int = 30  # Show Warnings of possible problems
    V_INFO: int = 20  # brief normal reporting
    V_DEBUG: int = 10  # detailed reporting for debugging

    # these next three are equivalent ... choice is good
    # logger messages will be forwarded with this setting
    # all other messages will be shown
    V_QUIET: int = 0
    V_NONE: int = 0
    V_STFU: int = 0


@dataclass(frozen=True)
class LogLevel:
    """ Logging Levels Numeric values.
        https://docs.python.org/3.9/library/logging.html#levels
        """

    LOG_CRITICAL: int = 50  # Serious error - unable to continue
    LOG_ERROR: int = 40  # Failure of essential function
    LOG_WARNING: int = 30  # Show Warnings of possible problems
    LOG_INFO: int = 20  # brief normal reporting
    LOG_DEBUG: int = 10  # detailed reporting for debugging

    # logger messages will be forwarded with this setting
    # all other messages will be shown
    LOG_NOTSET: int = 0


@dataclass
class DevConfig:
    """ Handle exceptions and logging for development. """

    debug: bool = True
    log_flag: bool = True
    re_raise: bool = True
    send_texts: bool = True
    send_emails: bool = True
    send_to_stderr: bool = True

    verbose: Verbosity = Verbosity.V_DEBUG
    log_level: LogLevel = LogLevel.LOG_DEBUG

    def __post_init__(self):
        if self.log_flag:
            self.logger = logging.getLogger(__name__)
        if self.send_texts:
            self.tm = TextMessaging()

    def log(e: Exception):
        """ Log an exception based on configuration. """
        self.logger()

    def e(e: Exception):
        """ Handle exceptions based on configuration. """
        pass

    def dev(e: Exception):
        pass


dev = DevConfig()


class PathFinder(Path):
    _flavour = _windows_flavour if os.name == "nt" else _posix_flavour

    def __init__(self, filename):
        super().__init__()
        self._path = Path(filename).resolve()
        print(self._path)
        self._filename = self._path.name
        self._create_file()
        print(self._filename)

        self._headers = []
        # _post_init allows the standard `init` to be left alone
        #   - cleaner for interited classes
        #   - handy for dataclasses
        self._post_init()

    def _post_init(self):
        pass

    def _create_file(self):
        if self.exists():
            print("exists")
        else:
            print("does not exist ... creating")
            try:
                self.touch()
            except Exception as e:
                log(e)
                dev(e)

        print(self._filename)
        return self._filename

    def rename(self, name):
        self._name = name
        return self._name

    @property
    def name(self):
        if not self._name:
            self._name = Path(filename).basename()
        return self._name

    @property
    def get_cwd(self) -> Path:
        return self.cwd().resolve()

    @property
    def basename(self) -> str:
        _file = Path(filename).basename()


class MarkdownFile(PathFinder):
    def __init__(*args, **kwargs):
        self._headers = []
        super.__init__(*args, **kwargs)

    def headers():
        # path = pathlib.Path.cwd() / 'test.md'
        if not self._headers:
            with open(self, mode="r") as fid:
                self._headers = [
                    line.strip() for line in fid if line.startswith("#")
                ]
        return self._headers
        # print('\n'.join(headers))


class PythonFile(PathFinder):
    def headers():
        # path = pathlib.Path.cwd() / 'test.py'
        if not self._headers:
            with open(self, mode="r") as f:
                self._headers = [
                    line.strip() for line in f if line.startswith("#")
                ]
        return self._headers
        # print('\n'.join(headers))


if __name__ == "__main__":
    p = PathFinder("faketest")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import math
import sys

from sys import stdin

list_of_doubles = [math.sin(x) for x in range(10 * 1000 * 1000)]

for arg in sys.__dir__():
    print(arg)


def main():
    """
    CLI script main entry point.
    """
    arg = sys.argv[1]


if __name__ == "__main__":  # if script is loaded directly from CLI
    main()



'''
Examples of <head> and <tail> functions
based on Python2 tail example at:
https://docs.python.org/2/library/collections.html
'''

from collections import deque
from os import linesep as NL


def head(filename, n=5):
    'Return the first n lines of a file'
    with open(filename) as fd:
        return fd.read().splitlines()[:n]


def tail(filename, n=5):
    'Return the last n lines of a file'
    with open(filename) as fd:
        return deque(fd.read().splitlines(), n)


def _example():
    n = 10
    s = '-'*40
    print(s)
    print(f'--> head({n})')
    print(NL.join(head('tail.py', n)))
    print(s)
    print(f'--> tail({n}')
    print(s)
    print(NL.join(tail('tail.py', n)))
    print(s)


if False:
    _example()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
if True:  # !------------------------ Imports
    import sys

    # import platform
    import logging

    from sys import stdout, stderr, argv
    from enum import unique as _unique, IntEnum as _IntEnum

    # from dis import dis
    # from functools import lru_cache
    # from os import environ as ENV, linesep as NL
    # from pathlib import Path

    from autosys.defaults import *
    from autosys._version import *

    __all__ = [  # objects exported for 'import *'
        "ARGS",
        "br",
        "db_column_ruler",
        "dbprint",
        "Dict",
        "ENV",
        "err",
        "ErrC",
        "Error",
        "hr",
        "IS_64BITS",
        "List",
        "log_error",
        "logex",
        "lru_cache",
        "NL",
        "PLATFORM",
        "read",
        "show_dis",
        "show_version",
        "stderr",
        "stdout",
        "SUPPORTS_COLOR",
        "Terminal",
        "tryit",
        "v_name",
        "verbose",
        "vprint",
    ]
if True:  # !------------------------ CONSTANTS

    if _log_flag_:
        _LOG_PATH: Path = Path().cwd().resolve() / "log_autosys.log"
        # logging.basicConfig(filename='example.log',
        #                     filemode='w', level=logging.DEBUG)
        logging.basicConfig(filename=_LOG_PATH.as_posix(), level=logging.DEBUG)
        logging.debug(f"Logging is on for {__title__} version {__version__}")
        logging.info(f"  {__copyright__}  -  License: {__license__}")

    # sometimes useful CONSTANTS ...
    # more reliable - https://docs.python.org/3.9/library/platform.html

    class LogSystem:
        def __init__():
            pass


if True:  # !------------------------ Logging Utilities

    def info(*args):
        logging.info(*args)

    def log_error(*args):
        """ Error reporting in debug mode.

            current (temporary) behavior: print error messages to <stderr> using dbprint. (if _debug_ flag is True)
            """
        # TODO - fix temporary functionality
        if _log_flag_:
            logging.info(*args)
            dbprint(*args)  # ! temp

    def logex(func):
        """ Decorator to catch and log errors. """
        if _log_flag_:
            try:
                result = func()
            except IOError as e:
                log_error(
                    f"logex caused an error while reporting <{func}>: {e}."
                )
            else:
                if isinstance(result, Exception):
                    log_error(f"Function <{func}> caught an error: {e}.")
            finally:
                del e

    def verbose(v: int, *args, **kwargs):
        """ Print based on '_verbose_' allowed verbosity level.

            v: int - requested verbosity level
            """
        try:  # if _verbose_ constant does not exist, skip this function
            _verbose_ == 0
        except NameError:
            return -1
        if _verbose_ >= v:
            if v < 2:
                kwargs["file"] = "sys.stdout"
                print(*args, **kwargs)
            elif v == 2:
                kwargs["file"] = "sys.stderr"
                print(*args, **kwargs)


if True:  # !------------------------ Script Tests

    def _tests_():
        """ Debug Tests for script. """

        pass


def _main_():  # !------------------------ Script Main
    """
    CLI script main entry point.
    """
    if _debug_:
        _tests_()


if __name__ == "__main__":
    _debug_: bool = True  # True => use Debug features
    _verbose_: int = 0  # Verbosity level => 0 - 4
    _log_flag_: bool = _debug_ and True  # True => log to file if _debug_
    # argv.append('version')  # ! test
    # argv.append('help')  # ! test
    term = Terminal()
    if "version" in ARGS:
        show_version()
    else:
        if "help" in ARGS or "debug" in ARGS:
            _debug_ = True
        _main_()

""" Error alternates:

    class _Enum_ish(tuple):
        # alternate Enum:
        # ref: https://stackoverflow.com/a/9201329
        __getattr__ = tuple.index
        __setattr__ = None
        __del__ = None


    class _Enum_like(object):
        # alternate Enum
        # ref: https://stackoverflow.com/a/4092436
        values = []

        class __metaclass__(type):
            def __getattr__(self, name):
                return self.values.index(name)

            def name_of(self, i):
                # There's another handy advantage: really fast reverse lookups:
                return self.values[i]

    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# shellcheck source=/dev/null
# shellcheck disable=2230,2086
f""" Part of the `AutoSys` package - utilities for macOS apps

        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`

    Usage: {__file__} [-h|-v] FILE

    """
if True:  # ?----------------------------------- Imports
    import time
    from typing import Dict, Generator, List
    from functools import lru_cache as lru_cache

    # import timeit
    from time import perf_counter_ns as timer
    from sys import argv, stderr, stdout
    from os import linesep as NL

    _debug_: bool = True


def dbprint(*args, sep=" ", end=NL, file=stderr, flush=False):
    if _debug_:
        print(
            *args, sep=sep, end=end, file=file, flush=flush,
        )


_FIB_FMT: str = "f'{func.__name__:>20.20}({args[0]}) = {result:>5d}   -  {dt:>10d} ns'"

time_list = []


def time_me(func, *args, **kwargs):
    def _timed(func, *args, **kwargs):
        dt = timer()
        result = func(*args, **kwargs)
        dt = timer() - dt

        dbprint("test")
        dbprint(f"{eval(set_fmt())}")
        return (dt, result)

    return _timed


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        dt = (te - ts) * 1000

        time_list.append({method.__name__: f"{dt:4.4}"})
        dbprint(time_list[-1])
        return (result, dt)

    return timed


def _fib(n: int):
    """ Calculates and returns the steps of the Fibonacci Sequence.

        This is the most simple and direct method.

        return:
            int - Fibonacci of <n>
        """
    return _fib(n - 1) + _fib(n - 2) if n > 1 else n


@lru_cache
def _fib_cache(n: int):
    """ Calculates and returns the steps of the Fibonacci Sequence.

        This is the most simple and direct method.

        return:
            int - Fibonacci of <n>
        """
    return _fib(n - 1) + _fib(n - 2) if n > 1 else n


@timeit
def fib(n: int, count: int):
    retval = 0
    for i in range(count):
        result = _fib(n)
        # dbprint(f"'{fib.__name__}' completed {i} trials of fib({n}) = {result}")
        retval += result
    # dbprint(f"'{fib.__name__}' completed {count} trials of fib({n}) sum = {retval}")
    return retval


def repeat(func, *args, count=20, **kwargs):
    retval = 0
    for i in range(count):
        result = func(*args, **kwargs)[1]
        # dbprint(f"'{fib.__name__}' completed {i} trials of fib({n}) = {result}")
        retval += result
    dbprint(f"'{func.__name__}' completed {count} repeats.")
    return retval


# ?----------------------------------- Main Entry Point


def __main__(args=argv[1:]):
    """
    CLI script main entry point.
    """
    x = repeat(fib, 20, 20)


if __name__ == "__main__":  # if script is loaded directly from CLI
    __main__()

# ?----------------------------------- References



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Tests for Autosys package. For more information:
https://docs.python-guide.org/writing/tests/
"""

# 'package imports'
# from autosys.debug import *
from autosys.cli.debug import ARGS, NL, br, dbprint, hr

from typing import Iterable

_debug_: bool = True  # True => use Debug features
__all__ = ["Lister"]


class Lister(list):
    DEV_BLACKLIST = ("_", "arepl")
    """ Utility to return various formats of an iterable. """

    # TODO - add textwrap functionality

    def __init__(self, iterable):
        super().__init__(iterable)
        self.data = list(iterable)

    def __all__(self):
        return [f"{x}" for x in self.filter(blacklist=Lister.DEV_BLACKLIST)]

    def __str__(self):
        return self.comma

    def join(self, sep=" ") -> (str):
        return sep.join(self)

    @property
    def lines(self, wrap=79) -> (str):
        return NL.join(self)

    @property
    def comma(self) -> (str):
        return ", ".join(self)

    @property
    def sort(self, rev=False):
        return Lister(sorted(self._list, reverse=rev))

    @property
    def len(self) -> (int):
        try:
            return len(self)
        except IndexError:
            return 0

    @property
    def size(self):
        return Lister(self._list)

    # @property
    # def max(self):
    #     return self._list.max()

    def bl(self, blacklist=DEV_BLACKLIST):
        return Lister([x for x in self._list if not x.startswith(blacklist)])

    def wl(self, whitelist=("")):
        return Lister([x for x in self._list if x.startswith(whitelist)])

    def filter(self, blacklist=DEV_BLACKLIST, whitelist=("")):
        """ Return the list containing only `whitelist` items that are not `blacklist` items.

            blacklist - items that are tossed out. defaults:
                - '__' : 'dundermethods'
                - '_' : 'dunders' and private variables
                - 'arepl' : (dev) from the AREPL VsCode Extension

            whitelist - items that are retained if not blacklisted. defaults:
                - '' : default to return all

            Notes:
                - '' as item in whitelist returns all (use as default)
                - () as blacklist returns all (use as default)

                - '' as item in blacklist returns nothing (edge case)
                - () as whitelist returns nothing (edge case)
            """
        # testing for these edge cases likely imposes significant performance
        # costs, especially the blacklist check. That one is avoided.
        if whitelist == ():
            return self.bl(blacklist=blacklist)
        # if '' in blacklist:
        #     return self.wl(whitelist=whitelist)
        return Lister(
            list([
                x for x in self
                if x.startswith(whitelist) and not x.startswith(blacklist)
            ]))


_lister = Lister(dir())
__all__ = Lister(dir()).__all__()
# print(__all__)


def _tests_():
    """ Debug Tests for script. """
    br()
    d = _lister

    # print(d.max)
    # hr()
    # print(d.len)
    # hr()
    print(d.join())
    # br()
    # print(d.comma)
    br()
    # print(d.wl())
    print(d.filter())
    print(d.__all__())
    print(d.lines)
    br()
    del d


def _main_():  # !------------------------ Script Main
    """
    CLI script main entry point.
    """
    if _debug_:
        _tests_()


if __name__ == "__main__":
    _main_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# !---------------------------------------------- Imports
if True:  # ! -- System Imports
    import requests
    import sys

if True:  # ! -- Specific System Imports
    from base64 import b64decode, b64encode
    from pathlib import Path
    from typing import Dict, List

    try:
        import ujson as json
    except:
        import json
if True:  # ! -- Package Imports
    from autosys.utils import dbprint, read, NL

#!---------------------------------------------- CONSTANTS
DEFAULT_BASE_URL: str = "https://api.twitter.com/"
CWD: Path = Path(__file__).resolve().parents[0]
DEFAULT_TWITTER_CREDENTIAL_FILE: Path = CWD / "twitter_credentials_private.json"
_debug_: bool = True

#!---------------------------------------------- Utilities


def try_it(func):
    """ Wrapper to catch exceptions.

        func   - function wrapped by decorator @try_it

        Return - normal function return or Exception
        """

    def tried(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return e

    return tried(*args, **kwargs)


def GET(url="http://www.example.com"):
    return requests.get(url)


def api_test():
    url = "http://httpbin.org/post"

    payload = '{\n    "key1": 1,\n    "key2": "value2"\n}'
    headers = {
        "Content-Type": "application/json,text/plain",
        "User-Agent": "PostmanRuntime/7.15.0",
        "Accept": "*/*",
        "Cache-Control": "no-cache",
        "Postman-Token": "e908a437-88ea-4b00-af53-7a9a49033830,ba90e008-0f7f-4576-beb8-b7739c8961f1",
        "Host": "httpbin.org",
        "accept-encoding": "gzip, deflate",
        "content-length": "42",
        "Connection": "keep-alive",
        "cache-control": "no-cache",
    }

    response = requests.request("POST", url, data=payload, headers=headers)

    print(response.text)


#!---------------------------------------------- encoding utilities


def encode64(s):
    """ Return ascii string encoded to Base64.

        utf-8 -> ascii -> base64 -> ascii """
    return b64encode(s.encode("ascii")).decode("ascii")


def decode64(s):
    """ Return ascii string decoded from Base64.

        utf-8 -> base64 -> ascii -> utf-8 """
    return b64decode(s.encode("ascii")).decode("ascii")


def binary_encode_64(binary_filename: str):
    """ Returns a base64 string from decoded binary file. """
    try:
        with open(binary_filename, "rb") as binary_file:
            return b64encode(binary_file.read()).decode("utf-8")
    except OSError as e:
        return ""


def binary_decode_64(base64_img: str, binary_filename: str):
    """ Encodes a base64 string to bytes and writes it to a binary file. """
    try:
        base64_img_bytes = base64_img.encode("utf-8")
        with open(binary_filename, "wb") as file_to_save:
            file_to_save.write(base64.decodebytes(base64_img_bytes))
        return 0  # 0 = success ...
    except OSError as e:
        return e


def print_binary(data):
    for byte in data:
        print(format(byte, "08b"), end=" ")


#!---------------------------------------------- Classes


class TwitterAPIError(Exception):
    """ Base Class for Twitter API Errors. """

    pass


class TwitterKeys:
    def __init__(self, filename=DEFAULT_TWITTER_CREDENTIAL_FILE):
        self._filename: Path = filename
        self._credentials: Dict[str, str] = {}
        self._load()

    def __str__(self):
        return self.encoded_key

    def _load(self):
        """ Get twitter credentials from file object. """
        self._credentials = {}
        self._encoded_key: Unicode = r""
        try:
            with open(self._filename, "r") as f:
                self._credentials = json.load(f)
        except Exception as e:
            raise TwitterAPIError(
                f"The credential file ({self._filename}) was not loaded.{NL}Error: {e}."
            )

    def save(self, credentials):
        """ Save the credentials dictionary object to a json file. """
        if isinstance(credentials, dict):
            self._credentials = credentials
            if self.check:
                with open(self._filename, "w") as f:
                    json.dump(credentials, f)
                self._load()  # ! read back file from disk to verify
                if not self.check:
                    raise TwitterAPIError(
                        f"The file did not load and verify correctly: {self._filename}"
                    )
                return 0
            raise TwitterAPIError(
                "The provided credentials are missing values."
            )
        raise TwitterAPIError(
            "The provided credentials are not in proper dictionary format."
        )

    @property
    def encoded_key(self):
        """ Create and store encoded key for request headers. """
        # ? should this be stored? cached? or created as needed each time?
        if not self._encoded_key:  # create key if needed
            self._encoded_key = encode64(f"{self.c_key}:{self.c_secret}")
        return self._encoded_key

    #!-------------------------------- TwitterKeys properties

    @property
    def check(self):
        """ Return 0 for good self-test, 1 otherwise. """
        return (
            self.c_key is None
            and self.c_secret is None
            and self.a_token is None
            and self.a_secret is None
        )

    @property
    def c_key(self):
        try:
            return self._credentials["CONSUMER_KEY"]
        except:
            return None

    @property
    def c_secret(self):
        try:
            return self._credentials["CONSUMER_SECRET"]
        except:
            return None

    @property
    def a_token(self):
        try:
            return self._credentials["ACCESS_TOKEN"]
        except:
            return None

    @property
    def a_secret(self):
        try:
            return self._credentials["ACCESS_SECRET"]
        except:
            return None


tk = TwitterKeys()


def twitter_credential_json_file_create():
    # Enter your keys/secrets as strings in the following fields
    credentials = {}
    credentials["CONSUMER_KEY"] = ""
    credentials["CONSUMER_SECRET"] = ""
    credentials["ACCESS_TOKEN"] = ""
    credentials["ACCESS_SECRET"] = ""
    tk_save = TwitterKeys()
    tk_save.save(credentials)
    del tk_save


class APIconnect:
    def __init__(self, base_url=DEFAULT_BASE_URL):
        pass
        # create url for API request:

    # dbprint(f"{base_url=}")
    # auth_url = '{}oauth2/token'.format(base_url)
    # dbprint(f"{auth_url=}")
    # dbprint()
    # auth_headers = {
    #     'Authorization': 'Basic {}'.format(tk),
    #     'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8'
    # }
    # auth_data = {
    #     'grant_type': 'client_credentials'
    # }

    # auth_resp = requests.post(auth_url, headers=auth_headers, data=auth_data)
    # dbprint('Auth response: ', auth_resp.status_code)

    # return 0


#!---------------------------------------------- Script Tests


# __all__ = [
# 'DEFAULT_BASE_URL',
# 'DEFAULT_TWITTER_CREDENTIAL_FILE',
# 'TwitterAPIError',
# 'try_it',
# 'TwitterKeys',
# 'tk',
# 'key_save',
# 'APIconnect',
# 'dbprint',
# 'read'
# ]


def _tests_(args):
    """ Run Debug Tests for script if _debug_ = True. """
    print(GET())
    # from autosys.utils import print_dict
    tk_tmp = TwitterKeys()
    print(f"{_debug_=}")
    # print(f"{DEFAULT_BASE_URL=}")
    # print(f"{DEFAULT_TWITTER_CREDENTIAL_FILE=}")

    print("Partial encoded key from class TwitterKeys: ")
    print(tk_tmp.encoded_key[1:5])
    tk_save
    # print_dict(globals())
    # api_test()


def _main_(args):
    """ CLI script main entry point. """
    #! script testing
    if _debug_:
        return _tests_(args)
    return 0


if __name__ == "__main__":  # if script is loaded directly from CLI
    global script_name
    script_name = sys.argv[0]
    _main_(sys.argv[1:])


""" Original long method before pulling it out to a utility function:

    @property
    def encoded_key(self):
        ''' Create encoded key for request headers. '''
        if not self._encoded_key:
            # Reformat the keys and encode them
            key_secret = f"{self.c_key}:{self.c_secret}".encode('ascii')
            # Transform from bytes to bytes that can be printed
            b64_encoded_key = b64encode(key_secret)
            # Transform from bytes back into Unicode
            self._encoded_key = b64_encoded_key.decode('ascii')
        return self._encoded_key

        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" as_testing.py """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal

# from __future__ import absolute_import

# import importlib
# import logging
# import linecache
# import os
# import time
# import sys
# import inspect
from typing import Any, Dict, FrozenSet, List, Sequence, Tuple

# Constants
DEFAULT_DICT_DISPLAY_SEPARATOR: str = ": "
DEFAULT_CLI_DISPLAY_WIDTH: int = 80
DEFAULT_CLI_FIELD_PADDING: int = 15
DEFAULT_CLI_FIELD_MIN_PADDING: int = 10

# module variables
e: BaseException = None
result: Any = None

# class TestException(Exception):
# """
# Exception handler override
# """

# Reference: https://stackoverflow.com/questions/9823936/python-how-do-i-know-what-type-of-exception-occurred
# template = "An exception of type {0} occurred. Arguments:\n{1!r}"
# message = template.format(type(ex).__name__, ex.args)
# print message

# a = 1


def v_name(the_var: Any) -> str:
    """
    Return string containing name of the_var
    """
    try:
        result = [
            var_name
            for var_name, var_val in inspect.currentframe().f_back.f_locals.items()
            if var_val is the_var
        ][0]
        return result
    except IndexError as e:
        return ""


def name_var(the_var: Any) -> List[str]:
    """
    Return list containing name and value of the_var
    """
    return [
        [
            var_name
            for var_name, var_val in inspect.currentframe().f_back.f_locals.items()
            if var_val is the_var
        ][0],
        the_var,
    ]


def lineno():
    """
    Returns the current line number in our program.
    """
    # Reference: http://code.activestate.com/recipes/145297-grabbing-the-current-line-number-easily/
    print(
        "line",
        inspect.getframeinfo(inspect.currentframe()).lineno,
        "of file",
        inspect.getframeinfo(inspect.currentframe()).filename,
    )
    return inspect.currentframe().f_back.f_lineno


def py_ls(path_name: str = ".") -> Exception:
    """
    List relative and absolute paths of files in <path_name>.
    Parameter: path_name: str
    Return: None for success or Exception
    """
    # Reference: https://developers.google.com/edu/python/utilities
    # Example pulls filenames from a dir, prints their relative and absolute paths

    try:
        filenames = os.listdir(path_name)
    except OSError as e:
        return e
    for filename in filenames:
        print(filename)  # foo.txt
        try:
            # dir/foo.txt (relative to current dir)
            print(os.path.join(dir, filename))
        except OSError as e:
            return e
        try:
            # /home/nick/dir/foo.txt
            print(os.path.abspath(os.path.join(dir, filename)))
        except OSError as e:
            return e
    return None


def _add_dots(
    s: str, n: int, suffix: str = " ...", add_dots: bool = True
) -> str:
    """
    Truncate and return formatted string to fit <s> in <n> spaces
        add ' ...' if <add_dots> is True
    """
    str_len = n - len(suffix)
    if len(s) > str_len:
        s = s.ljust(str_len) + suffix
    return s


def get_module_sig(self, module, parameter_list):
    import inspect

    sig = inspect.signature(module)
    bound = sig.bind(
        "this is arg1",
        "this is arg2",
        "this is an extra positional argument",
        extra_named_arg="value",
    )
    print("Arguments:")

    for name, value in bound.arguments.items():
        print("{} = {!r}".format(name, value))
    # for name, data in inspect.getmembers(example):
    #     if name.startswith('__'):
    #         continue
    #     print('{} : {!r}'.format(name, data))


def print_var(
    the_string_name: str,
    the_string: str,
    p: int = DEFAULT_CLI_FIELD_PADDING,
    w: int = DEFAULT_CLI_DISPLAY_WIDTH,
    sep: str = DEFAULT_DICT_DISPLAY_SEPARATOR,
    print_it: bool = True,
) -> str:
    """ Format string for 'var : value' pattern
        string_tuple: Tuple[str, str] - (name, value) of variable
        p: int - padding; len of name field
        w: int - width ; len of return string
        sep: str - separator
        print_it: bool - print to CLI within function
        """
    str_name = [
        var_name
        for var_name, var_val in inspect.currentframe().f_back.f_locals.items()
        if var_val is the_string_name
    ][0]
    print("str_name: ", str_name)
    result: str = ""
    if p == 0:
        p = len(the_string_name)
    if p <= DEFAULT_CLI_FIELD_MIN_PADDING:
        p = DEFAULT_CLI_FIELD_MIN_PADDING

    str_padding: int = w - p - len(sep)
    # print('p: ', p, ' str_p: ', str_padding)
    str_format: str = f"{{:<{p}.{p}}}{sep}{{:<{str_padding}.{str_padding}}}"
    print(str_format)
    # print(_add_dots(the_string, 16))
    # result = str_format.format(
    #     _add_dots(the_string_name, p), _add_dots(the_string, str_padding)
    # )
    if print_it:
        print(result)
    # print()
    # print("str_format: ", str_format)
    # print("result: ", result)
    # print()
    return result


def get_module_info(module: str):
    return 0
    # try:
    #     (name, suffix, mode, mtype) = inspect.getmoduleinfo(filename)
    # except TypeError:
    #     print('Could not determine module type of %s' % filename)
    # else:
    #     mtype_name = {imp.PY_SOURCE: 'source',
    #                   imp.PY_COMPILED: 'compiled',
    #                   }.get(mtype, mtype)

    #     mode_description = {'rb': '(read-binary)',
    #                         'U': '(universal newline)',
    #                         }.get(mode, '')

    #     print('NAME   :', name)
    #     print('SUFFIX :', suffix)
    #     print('MODE   :', mode, mode_description)
    #     print('MTYPE  :', mtype_name)


def _pprint_globals():
    """
    Pretty Print all global variables.
    Designed for debugging purposes.
    """
    print()
    print("Globals: ")
    print("*" * 40)
    width = max(len(i) for i in globals())
    print(width)
    for s in globals():
        print("{:<15.15} : {:<64.64}".format(s, str(globals().get(s))))
    print()
    test_list: List[int] = [1, 2, 3]
    test_list.append(4)
    test_list.append("five")


def _pprint_dict_table(data: Dict[Any, Any], name: str = "Data Table"):
    """
    Pretty Print dictionary in table format.
    """
    print()
    print(name.title())
    print("*" * len(name))
    width = max(len(i) for i in data)
    print(width)
    for s in globals():
        print("{:<15.15} : {:<64.64}".format(s, str(globals().get(s))))
    print()
    test_list: List[int] = [1, 2, 3]
    test_list.append(4)
    test_list.append("five")


def _execute_test_code(tests: List[str]) -> List[Exception]:
    result: List[Exception] = []
    for test in tests:
        try:
            # run test code
            print("=> ", test)
            exec(test)
        except Exception as e:
            result.append(e)
    return result


def _run_tests() -> str:
    tests: List[str] = [
        "_pprint_globals()",
        'print("basename: ", basename(__file__))',
        "print(1/0)",
    ]
    _execute_test_code(tests)
    # result = '\n'.join(str(e) for e in _execute_test_code(tests))
    # if result:
    #     print(result)
    # return result
    # print("traceback: ", traceback.format_exc())
    print("e.__class__.__name__", e.__class__.__name__)
    # log.exception(e)
    print("e: ", e)
    print("e.args: ", e.args)
    print("type(e): ", type(e))


def _pprint_code_tests(tests: List[Exception]):
    pass


def _execute_test_code(tests: List[str]) -> List[Exception]:
    """
    Execute and report exceptions for code snippets.
    """
    result: List[Exception] = []
    for test in tests:
        try:
            # run test code
            print("=> ", test)
            exec(test)
        except Exception as e:
            result.append(e)
    return result


def _run_tests() -> str:
    """
    Run and report on specific tests for some functions in this module.
    """
    tests: List[str] = [
        "_pprint_globals()",
        'print("basename: ", basename(__file__))',
        "print(1/0)",
    ]

    for e in _execute_test_code(tests):
        # result = '\n'.join(str(e) for e in _execute_test_code(tests))
        # if result:
        #     print(result)
        # return result
        # print("traceback: ", traceback.format_exc())
        print("e.__class__.__name__", e.__class__.__name__)
        # log.exception(e)
        print("e: ", e)
        print("e.args: ", e.args)
        print("type(e): ", type(e))


def _pprint_code_tests(tests: List[Exception]):
    pass


if __name__ == "__main__":

    if len(sys.argv) >= 2:
        filename = sys.argv[1]
    else:
        filename = "example.py"

    print()
    print("Debug print values")
    print("*" * 80)
    print()
    # print('_add_dots(name, 25, 15) : ', _add_dots(name, 25))
    # print('_add_dots(__name__, 25) : ', _add_dots(__name__, 25))
    # print('_add_dots(__file__, 45) : ', _add_dots(__file__, 45))
    # print('_add_dots(__version__, 15) : ', _add_dots(__version__, 15))
    # print('_add_dots(__package__, 25) : ', _add_dots(__package__, 25))
    print("v and name: ", name_var(__file__))
    print("get name - file: ", v_name(__file__))
    print("")
    # print_var()
    # print(locals())
    widths = [len(key) for key in list(locals().keys())]
    keys = [key for key in list(locals().keys())]
    padding = max(widths)
    # print(widths)
    # print()
    # print(keys)
    # for key in keys:
    #     # print(key, locals().get(key))
    #     print_var(key, locals().get(key))
    print()
    # get_module_sig(inspect, )
    # log = logging.getLogger()
    print(_run_tests())
    print()
    print("... Tests Complete.")
    _pprint_dict_table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Common imports, CONSTANTS, and utilities for the `AutoSys` package

    Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'Standard Library'
import dataclasses
import json
import os
import string

from dataclasses import Field, dataclass, field
from os import linesep as NL

from typing import Dict, List

NUL: str = chr(0)

STR_ALPHA: str = string.ascii_letters
STR_ALPHANUMERIC: str = string.ascii_letters + string.digits
STR_HEX: str = string.hexdigits
STR_NAMES: str = string.ascii_letters + " _-"
STR_WHITESPACE: str = string.whitespace
STR_PRINTABLE: str = string.printable
STR_PUNCTUATION: str = string.punctuation



#!/usr/bin/env python3
""" #### AutoSys Setup

    ```
    Usage:  Setup [-dv] [--setname=<name>] [--setver=<version>] <setup_args>
            Setup [-h | --help] [--version]

    Arguments:
        setup_args              passed through to Python Setup (*)

    Options:
        --setname=<name>        Set Package Name
        --setver=<version>      Set Version Name
        -d --debug              Set Debug Mode [default=True]
        -v --verbose            Set Verbose Mode [default=True]
        -h --help               Show this screen.
        --version               Show version.

    * Use single quotes around 'setup_args' that contain dashes in order to
    pass them through to the Python Setup utility.

    ```
    e.g. Use these:
    ```
    python3 -m setup 'build -vn'
    python3 -m setup '--help-commands'
    ```
    instead of these:
    ```
    python3 -m setup build -vn
    python3 -m setup --help-commands
    ```
    """
""" Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """
# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine
# flake8: noqa: E501

_debug_: bool = False

# ? ################################### Default Metadata
NAME: str = "AutoSys"
USERNAME: str = "skeptycal"
DESCRIPTION: str = "System utilities for Python on macOS."
PYTHON_REQUIRES: str = ">=3.6.0"
# ? ####################################################

if True:  # ? ################### imports
    import os
    import re
    import sys

    from io import TextIOWrapper
    from os import linesep as NL
    from sys import stderr, path as PYTHONPATH

    IS_WIN32 = sys.platform == "win32" or (getattr(os, "_name", False) == "nt")

    if sys.version_info[:2] >= (3, 6):
        from pathlib import Path, PurePath
    else:
        from pathlib2 import Path, PurePath

    try:
        from locale import getpreferredencoding

        DEFAULT_ENCODING = getpreferredencoding(do_setlocale=True)
        del getpreferredencoding
    except ImportError:
        DEFAULT_ENCODING = "utf-8"
    except Exception:
        DEFAULT_ENCODING = "utf-8"
        del getpreferredencoding

    # from loguru import logger
    from auto_loguru import logger
    from setuptools import find_namespace_packages, setup

    # from pep517.envbuild import build_wheel, build_sdist

    from typing import (
        Any,
        AnyStr,
        Dict,
        List,
        Match,
        Optional,
        Sequence,
        Set,
        Tuple,
        Union,
    )

    PathLike = Union[Path, str, None]


if True:  # ? ################### imports
    PY3: bool = sys.version_info.major > 2

    HOME: str = Path().home().as_posix()
    HERE: str = Path(__file__).resolve().parent.as_posix()

    if not NAME:
        NAME = Path(HERE).name

    LOG_FORMAT: str = "{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}"
    # choose a common location if you consolidate logs
    LOG_LOCATION: str = f"{HERE}/logs/"

    def readme(file_name: str = "readme.md", search_list: List[str] = []) -> str:
        """ Returns the text of the file (defaults to README files)

            The default file is `README.md` and is *NOT* case sensitive.
            (e.g. `README` is the same as `readme`)
            Can load *any* text file, but the default search path is setup
            for readme files

            ```
            Search path = ["readme.md", "readme.rst", "readme", "readme.txt"]
            ```

            Example:

            ```
            long_description=readme()
            ```
            """

        # add default search list for README files
        if not search_list:
            search_list = ["readme.md", "readme.rst", "readme", "readme.txt"]
        # make sure 'file_name' is in 'search_list' at index 0
        if file_name not in search_list:
            search_list.insert(0, file_name)
        found: bool = False
        # traverse up through directory tree searching for each
        # file in 'search_list'
        for searchfile in search_list:
            # search in this script's path and above
            for parent in Path(file_name).resolve().parents:
                find_path = Path(parent / searchfile)
                if find_path.exists():
                    found = True
                    break
            if found:
                break
        if found:
            try:
                with open(find_path, mode="r", encoding=DEFAULT_ENCODING) as f:
                    return f.read()
            except OSError:
                raise OSError(
                    f"Cannot read from the \
                    'readme' file '{find_path}'",
                )
        else:
            raise FileNotFoundError(
                f"Cannot find project 'readme' file in project \
                    tree. Search list = {search_list}",
            )

    def pip_safe_name(s: str) -> str:
        """ Return a name that is converted to pypi safe format.
            ####
            (Returns a lowercase string has no spaces or dashes)
            """
        return s.lower().replace("-", "_").replace(" ", "_")

    NAME = pip_safe_name(NAME)

    def get_version(file_name: str = f"{HERE}/pyproject.toml") -> str:
        logger.info(f"pyproject.toml file: {file_name}")
        with Path(file_name).open("r") as fp:
            regex_version: re.Pattern[Any] = re.compile(
                r'^version\s*=\s*[\'"]([^\'"]*)[\'"]', re.MULTILINE,
            )
            return re.search(regex_version, fp.read()).group(1)  # type: ignore

    VERSION: str = get_version()
    __version__: str = VERSION
    VERSION_INFO: Tuple[int, ...] = tuple(map(int, VERSION.split(".")))

    def table_print(data: (Dict, Sequence), **kwargs) -> None:  # type: ignore
        """ Pretty Print sequences or dictionaries.
        """
        tmp: List[str] = []
        if isinstance(data, AnyStr):
            raise TypeError(
                'Strings and bytes cannot be used for "table_print". \
                    Try another form of iterable',
            )

        elif isinstance(data, dict):
            key_width: int = len(max(data.keys()))
            print(f"key_width = {key_width}.")
            tmp.extend(
                [
                    f"{str(k):<15.15} :  \
                    {repr(v):<45.45}"
                    for k, v in data.items()
                ],
            )
        elif isinstance(data, (list, tuple, set)):
            for x in data:
                try:
                    tmp.append(f"{str(x):<15.15} :  {repr(f'{x}'):<45.45}")
                except:
                    tmp.append(f"{str(x)}")
        else:
            raise TypeError(
                "Parameter must be an iterable Mapping or Sequence \
                    (Strings are excluded).",
            )
        print(NL.join(tmp), **kwargs)


class LoggedException(Exception):
    def __init__(self, *args: Any) -> None:
        logger.error(*args)
        super().__init__(*args)


@logger.catch
class SetupConfigError(LoggedException):
    """ An error occurred in SetupConfig. """


@logger.catch
class SetupConfig:
    def __init__(
        self,
        name: str = NAME,
        version: str = VERSION,
        description: str = DESCRIPTION,
        python_requires: str = PYTHON_REQUIRES,
        debug: bool = False,
        verbose: bool = False,
        logging: bool = True,
        encoding: str = DEFAULT_ENCODING,
    ):

        self._debug_: bool = debug
        self._verbose_: bool = verbose
        self._logging_: bool = logging
        if self.logging:  # TODO - check this stuff
            # standard log - kept for 3 days
            logger.add(f"{LOG_LOCATION}{NAME}.log", retention="3 days")
            # json log - kept for 10 days
            logger.add(
                f"{LOG_LOCATION}{NAME}_log.json", serialize=True, retention="10 days",
            )
        if self.debug:
            logger.configure()
            logger.level = "DEBUG"  # TODO - level setting is not working
        elif self.verbose:
            logger.level = "INFO"
        else:
            logger.level = "SUCCESS"
        logger.debug(f"Debug value is set to '{self.debug}'")
        logger.debug(f"Verbose value is set to '{self.verbose}'")

        self._name_: str = name
        logger.debug(f"Package name: {self.name}")
        self._version_: str = version
        logger.debug(f"Package version: '{self.version}'")
        self._here_: str = ""
        logger.debug(f"Script path: {self.here}")
        self._meta_data_: Dict[str, Any] = {}
        self._description_: str = description
        self._python_requires_: str = python_requires
        self._encoding_: str = encoding

    def dbprint(self, *args, **kwargs) -> None:  # type: ignore
        """ Print Debug Output to stderr. """
        if self.debug:
            print(*args, file=stderr, **kwargs)  # type: ignore

    def info(self, *args: Sequence[str]) -> None:
        """ Join args into a single string 'info' message and log it. """
        msg: str = ""
        # TODO - ' '.join(args) had issues
        for arg in args:
            # msg += str(arg) # TODO - is this better than fstrings? ...
            msg = f"{msg}{str(arg)} "
        logger.info(msg.rstrip())

    def err(self, e: Exception) -> None:
        """ Log last exception. """
        logger.exception(e)

    @property
    def name(self) -> str:
        if not self._name_:
            if NAME:
                self._name_ = pip_safe_name(NAME)
            else:
                self._name_ = pip_safe_name(Path(self.here).name)
        return self._name_

    @property
    def version(self) -> str:
        """ Return verbose flag. """
        if not self._version_:
            if VERSION:
                self._version_ = VERSION
            else:
                self._version_ = "0.0.1"
        return self._version_

    @property
    def python_requires(self) -> str:
        if not self._python_requires_:
            self._python_requires_ = PYTHON_REQUIRES
        return self._python_requires_

    @property
    def description(self) -> str:
        if not self._description_:
            self._description_ = DESCRIPTION
        return self._description_

    @property
    def here(self) -> str:
        if not self._here_:
            self._here_ = Path(__file__).resolve().parent.as_posix()
        return self._here_

    @property
    def debug(self) -> bool:
        """ Return debug flag. """
        if not self._debug_:
            self._debug_ = _debug_
        return self._debug_

    @property
    def logging(self) -> bool:
        """ Return debug flag. """
        if not self._logging_:
            self._logging_ = True
        return self._logging_

    @property
    def verbose(self) -> bool:
        """ Return verbose flag. """
        if not self._verbose_:
            self._verbose_ = True
        return self._verbose_

    @property
    def encoding(self) -> str:
        """ Return file encoding (or lazy load default) """
        if not self._encoding_:
            self._encoding_ = DEFAULT_ENCODING
        return self._encoding_

    def run(self) -> None:
        if self.debug:
            table_print(self.__dict__)
        else:
            setup(
                name=NAME,
                version=VERSION,
                long_description=readme(),
                package_dir={"": f"{self.name}"},
                packages=find_namespace_packages(
                    f"{self.name}", exclude=["*test*", "*bak*"],
                ),
                py_modules=[f"{self.name}"],
                url=f"https://{USERNAME}.github.io/{self.name}/",
                download_url=f"https://github.com/{USERNAME}/{self.name}/archive/{self.version}.tar.gz",
                project_urls={
                    "Website": f"https://{USERNAME}.github.io/{self.name}/",
                    "Documentation": f"https://{USERNAME}.github.io/{self.name}/docs",
                    "Source Code": f"https://www.github.com/{USERNAME}/{self.name}/",
                    "Changelog": f"https://github.com/{USERNAME}/{self.name}/blob/master/CHANGELOG.md",
                },
            )


# ? ############################## Setup!
sc = SetupConfig(debug=_debug_)


@logger.catch
def main() -> None:
    sc.info(f"SetupConfig for {sc.name} version {sc.version}.")
    if sc.debug:  # do some live tests in debug mode ...
        logger.debug(f"sc.debug mode set to {sc.debug}.")
    else:  # run setup ...
        sc.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" _version.py - version and demographic information

    Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'Standard Library'
from io import TextIOWrapper
from locale import getpreferredencoding
from logging import Logger
from pathlib import Path
from sys import argv
from tempfile import NamedTemporaryFile

from typing import Dict, Final, List, Tuple


class FileSet(list):
    input_list: List[str]
    temp_file: TextIOWrapper
    log: Logger

    def __init__(self, *args):
        self.input_list = args
        self.temp_file = NamedTemporaryFile(mode="wt", prefix=__file__)
        self.log = Logger(__file__)

    def __str__(self):
        return self.lines

    def txt_replace(self, needle, haystack):
        pass

    def drop_lines(self):
        for original in files:
            with NamedTemporaryFile(mode="wt", prefix=__file__,) as temp_file:
                with open(
                    file=original, mode="rt", encoding=self.DEFAULT_ENCODING
                ) as fd:
                    data = fd.readlines()

                # process lines

                with open(file=f""):
                    pass

                Path(temp_filename).replace(original_filename)



# 'Standard Library'
import sys


def public(f):
    """"Use a decorator to avoid retyping function/class names.

        * Blog post by Sam Denton:
        http://code.activestate.com/recipes/576993-public-decorator-adds-an-item-to-__all__/

        * Based on an idea by Duncan Booth:
        http://groups.google.com/group/comp.lang.python/msg/11cbb03e09611b8a

        * Improved via a suggestion by Dave Angel:
        http://groups.google.com/group/comp.lang.python/msg/3d400fb22d8a42e1
        """
    all = sys.modules[f.__module__].__dict__.setdefault("__all__", [])
    if f.__name__ not in all:  # Prevent duplicates if run from an IDE.
        all.append(f.__name__)
    return f


x: int = 2

public(public)  # Emulate decorating ourself
public(x)



timeit_results: List[float] = []


def timeit_init(arg: str = "init") -> bool:
    global timeit_results
    if arg == "init":
        timeit_results = []
    elif arg == "log_on":
        timeit_log = True
    elif arg == "log_off":
        timeit_log = False


class TimeIt(list):
    log: bool = False

    def __init__(self, log_flag: bool = True, print_flag: bool = True):
        self.log_flag = log_flag
        self.print_flag = print_flag

    def reset():
        self.clear()


def timeit(method):
    """
    Decorator - code timer for comparing and optimizing snippets
    """

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        timeit_results.append(te - ts)
        print("%r (%r, %r) %2.2f sec" % (method.__name__, args, kw, te - ts))
        return result

    return timed



# -*- coding: utf-8 -*-

# Reference: http://www.pythonlearn.com/html-009/book011.html
import string

fhand = open("romeo-full.txt")
counts = dict()
for line in fhand:
    line = line.translate(None, string.punctuation)
    line = line.lower()
    words = line.split()
    for word in words:
        if word not in counts:
            counts[word] = 1
        else:
            counts[word] += 1

# Sort the dictionary by value
lst = list()
for key, val in counts.items():
    lst.append((val, key))

lst.sort(reverse=True)

for key, val in lst[:10]:
    print key, val



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Creates twitter credentials json file from user authentication data.

#! Keep this file *PRIVATE* and do not submit it to source control!

Keys and info can be obtained from the twitter developer site:
https://developer.twitter.com/en (English Version, others available)

------------------------------------------------------------------------------

MIT License

    Copyright (c) 2018 - 2020 Michael Treanor <https://www.skeptycal.com>

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to
    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
    the Software, and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
    FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
    COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
    IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    """

# 'Standard Library'
import json

# Enter your keys/secrets as strings in the following fields
credentials = {}
credentials["CONSUMER_KEY"] = "<put your information here>"
credentials["CONSUMER_SECRET"] = "<put your information here>"
credentials["ACCESS_TOKEN"] = "<put your information here>"
credentials["ACCESS_SECRET"] = "<put your information here>"

# Save the credentials object to file
with open("twitter_credentials_private.json", "w") as f:
    json.dump(credentials, f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
# watching "Arrays vs Linked Lists - Computerphile"

# what about an array of linked elements?
# it is really a linked list, but will it have some advantages
#   of an array?

# during downtime they are copied back into array format
#   i.e. like memories, when they are sleeping, they are reorganized ...

# during busy times, inserting and removing creates either an extra
#   pointer to new elements or a temporary 'hole' element
#   the 'hole' elements are simply a 'nul' object that would
#   be automatically skipped

# this would be especially good when eleents are sometimes deleted
#   but rarely added

# if new elements are added often, this will struggle

# 'Standard Library'
import time

# 'package imports'
import autosys.lister

# profile with clock ticks as units ...


class Node:
    def __init__(self, data, next_element=None):
        self.data = data  # Assign data
        self.next = None  # Initialize next as null


class DoubleNode(Node):
    def __init__(self, data, prev_element=Element.last, next_element=None):
        super.__init__(self, data=data, next_element=next_element)
        self.last = prev_element  # maybe?


class CircularNode(DoubleNode):
    def __init__(
        self, data, prev_element=Element.last, next_element=None, head=None
    ):
        super.__init__(
            self, data, prev_element=prev_element, next_element=next_element
        )
        self.next = head  # if not head: self.next = None


class LinkedList:
    """
    The class constructor needs to make sure there is a 'first' element,
        track the final element, and initialize new elements with the 
        'last' as 'previous' and 'next' of 'last' as the new one. The
        'next' of new is null.
    """

    def __init__(self):
        self.head = None
        self.index = []  # TODO - just for testing ...

    # This function prints contents of linked list
    # starting from head

    def printList(self):
        temp = self.head
        while temp:
            print(temp.data, end=", ")
            temp = temp.next

    def append(self, node):
        pass

    def insert(self, node):
        pass

    def delete(self, node):
        pass


class DoubleLinkedList(LinkedList):
    pass


class CircularLinkedList(LinkedList):
    pass


# ------------------------------------------------------------------
# watching "Sorting Secret - Computerphile"

# The visual of the 'sorting box' is interesting ...

# what if we just subtract one from the other, e.g.


util = autosys.lister.Lister([])


def sort_subtract(a, b):
    diff = a - b  # this is the difference ... does this help?
    if diff > 0:
        return (b, a)
    return (a, b)


def sort_compare(a, b):
    if a > b:
        return (b, a)
    return (a, b)


# gradient indexing ... downtime maintaining a 'weighted average' index

# e.g. - with example data:
#   AAABCDEEEEEEEEFFFFFFFFGHIJKLLLLMMMMMMMMMMNNNNOPQRSSSSSSSSSSTTTTTTTTTUVWXYZ


def sort_ins(input_list):  # standard iterable or generator
    for x in input_list:
        pass


# builtins ...
def sort_numbers_in_place(input_list, case_sens=True):
    if case_sens:
        input_list = map(lower(), input_list)
        # map(lambda x: x.lower(), ["A", "B", "C"])
    return input_list.sort(reversed=False)


def sort_str_with_lowe(
    input_list, case_sens=True, rev_case=False, reverse=False
):
    if case_sens:
        pass  # just sort it ...


def sort_str_in_parts(
    input_list, case_sens=True, rev_case=False, reverse=False
):
    if case_sens:
        pass  # create 2 lists (upper, lower) and sort before recombining them


def sort_numbers_new_list(input_list):
    return sorted(input_list, reverse=False)


def call_all_sorts(test_lists):
    for L in test_lists:
        pass


test_lists = []
test_lists.append([5, 3, 2, 1, 4])
test_lists.append(["H", "j", "A", "A", "b", "x", "X"])
test_lists.append(["H", "j", "A", "A", "b", "x", "X", " "])
test_lists.append(["H", "j", "A", "A", "b", "x", "X", ""])
test_lists.append([5, 3, 2, 1, 4])
test_lists.append([5, 3, 2, 1, 4])
test_lists.append([5, 3, 2, 1, 4])


call_all_sorts(test_lists)


# using islower and isupper, paired with the format that is 'usual' for the
# data, will speed up the operation significantly
# https://www.dotnetperls.com/lower-python


def test_is_lower():
    """
    1384626116.650
    1384626122.033    lower():               5.38 s
    1384626124.027    islower() and lower(): 1.99 s
    """
    import time

    value = "intuitive"

    print(time.time())

    # Version 1: lower.
    i = 0
    while i < 10000000:
        v = value.lower()
        i += 1

    print(time.time())

    # Version 2: islower and lower.
    i = 0
    while i < 10000000:
        if not value.islower():
            v = value.lower()
        i += 1

    print(time.time())


test_is_lower()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import sys

from pathlib import Path

from typing import List

PYTHON_PATH = sys.path


def find_python_root(here: str):
    for d in Path(__file__).resolve().parents:
        if Path.exists((d / "README.md")):
            return d
    return None


def check_add_path(here: str):
    import sys

    here_path = Path(here).resolve().parents[0].as_posix()
    if not (here_path in sys.path):
        sys.path.append(here_path)


def listf(
    p: Path,
    pattern: str = "*",
    color: bool = True,
    dirs_first: bool = True,
    options: str = "lAhFp",
) -> List[str]:
    """ Generate file list for ls command.
        p - Path object to search
        pattern - pattern to match (default "*")
        color - use color (default True)
        sort - priority order in [none, alpha, size, owner, classify, type]

        bonus: A Brief History of the 'ls' command
        http://www.tldp.org/LDP/LG/issue48/fischer.html
        """
    for d in p.glob(pattern):
        if d.is_dir():
            details = d.exists() * "*"
            details += d.is_dir() * "/"
            details += d.is_file() * "f"
            details += d.is_symlink() * "l"
            name = d.stem
            ext = d.suffix
            fmt = 'f" {details:<6.6} {name:<15.15} {ext:<5.5}"'
            print(f"{eval(fmt)}")


def ls(p: str = "."):
    print("-" * 50)
    print(f"Path contents of {str(p):<25}")
    print("-" * 50)
    print(listf(Path(p)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" readme.py

    locate and return the project's readme file

    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

# 'Standard Library'
from pathlib import Path

try:
    DEFAULT_ENCODING
except:
    from locale import getpreferredencoding

    DEFAULT_ENCODING: str = getpreferredencoding(do_setlocale=True) or "utf-8"


def readme(filename: str = "readme.md"):
    """ Returns the text of the README file

        The default file is `README.md` and is *NOT* case sensitive. (e.g. `README` is the same as `readme`)
        Can load *any* text file, but the default search path is setup for readme files

        ```
        Search path = ["readme.md", "readme.rst", "readme", "readme.txt"]
        ```

        Example:

        ```
        long_description=readme()
        ```
        """

    search_list = ["readme.md", "readme.rst", "readme", "readme.txt"]
    if filename not in search_list:
        search_list.insert(0, filename)
    found: bool = False
    for searchfile in search_list:
        for parent in Path(__file__).resolve().parents:
            find_path = Path(parent / searchfile)
            if find_path.exists():
                found = True
                print(find_path)
                break
        if found:
            break
    if found:
        try:
            with open(find_path, mode="r", encoding=DEFAULT_ENCODING) as f:
                return f.read()
        except IOError as e:
            return e
    else:
        raise FileNotFoundError(
            f"Cannot find project 'readme' file in project tree. Search list = {search_list}"
        )



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import sys
import time

from collections import deque

from typing import Deque, List  # , Iterator,

# TODO interesting features of patterns of numbers
# when looking for primes, you first do the easy things:
# get rid of even numbers ... well, is there a way to easily
#   and 'cheaply' get rid of numbers divisible by 3?
#   maybe keeping track of the most recent number divisible by 3?
#   maybe ... counting by 3's as you proceed through the loop?
#                      1111111111222222222223333333334
#            01234567890123456789012345678901234567890
#   evens      2 x x x x x x x x x x x x x x x x x x x
#   threes      3  .  x  .  x  .  x  .  x  .  x  .  x
#   fives        45    .    .    .    x    .    x    .
#   ... which I guess amounts to skipping every 6th number?
#   fours are accounted for ...
#   fives? since the even numbers are already gone ... this amounts to
#   figuring out which numbers are divisible by 5, starting at 5 and ...
#   counting by 10's
#   maybe we can take this number, change it to base 3 instead of 2 ...
#   and look to the low bit, just like the binary odd/even test?
#
#   how hard is it to change numbers to prime number bases and count by <base>
#   I bet there are a great many numbers that can be eliminated this way
#   maybe some other way?

# ---------------------------- Utilities


def var_name(*args):
    for arg in args:
        return f"{arg.__name__}"


def timeit(method):
    def timed(*args, **kw):
        s0 = sys.getsizeof(method)
        t0 = time.time()
        result = method(*args, **kw)
        dt = time.time() - t0
        s1 = sys.getsizeof(method)
        print(method.__name__)
        fn = var_name(kw.get("func"))
        # print(func_name)

        # if 'log_time' in kw:
        #     name = kw.get('log_name', method.__name__.upper())
        #     kw['log_time'][name] = int((dt) * 1000)
        # else:
        print(f"{fn:25.25} - {dt*1000:>6.6} \t{'ms':>3.2}")
        return result

    return timed


# ---------------------------- Test Functions
if True:

    def fibonacci(n: int = 2) -> int:
        if n == 0 or n == 1:
            return n
        else:
            return fibonacci(n - 1) + fibonacci(n - 2)

    def fibonacci_2(n: int = 2) -> int:
        if n > 1:
            return fibonacci(n - 1) + fibonacci(n - 2)
        else:
            return n

    def fibomemo(n: int = 2) -> int:
        memo: List[int] = [0, 1]
        for i in range(2, n + 1):
            memo.append(memo[i - 1] + memo[i - 2])
        return memo[n]

    def fibo_deque(n: int = 2) -> int:
        memo: Deque[int] = deque([0, 1], 2)
        for i in range(2, n + 1):
            memo.append(memo[0] + memo[1])
        return memo[-1]

    @timeit
    def fibo_fake_deque(n: int = 2) -> int:
        v1: int = 0
        v2: int = 1
        v3: int = 0
        for i in range(2, n + 1):
            v3 = v1 + v2
            v1 = v2
            v2 = v3
        return v3

    def fibo_fake_deque_2(n: int = 2) -> int:
        v1: int = 0
        v2: int = 1
        v3: int = 0
        for i in range(2, n + 1):
            v3 = v1 + v2
            v1 = v2
            v2 = v3
        return v3

    def fibo_fake_deque_3(n: int = 2) -> int:
        v1: int = 0
        v2: int = 1
        v3: int = 0
        for i in range(2, n + 1):
            v3 = v1 + v2
            v1 = v2
            v2 = v3
        return v3

    # ---------------------------- Caller Functions


@timeit
class Profiler_(list):
    DEFAULT_REPS = 10
    DEFAULT_DURATION = 100
    DEFAULT_MAX_DURATION = 1000

    class Code:
        """ Objects containing code blocks and/or data for a function.

            Code objects contain the following:

            - code           : string or existing function
            - description    : short description used for reporting
            - args           : list of arguments required by the function
            """

        def __init__(self, code, description, args):
            super().__init__()
            self.code = code
            self.description = description
            self.args = Profiler_.arg_list(args)

        def function(self):
            """ Return function with arguments ready to call. """
            return f"{self.code.__name__}({self.args})"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # add external functions
        self.clear()

    def add(self, code):
        """ Add a Code object to the profiler list.

            - create Code object
            - add item to test list
            """
        if type(code) == "function":
            c = Code(code, code.__doc__, args)
        if type(code) == str:
            pass
        self.append()

    def add_internal(self, func_list):
        """ Add a list of Code objects to the profiler list.

            - create Code objects from function list
            - add Codes to Profiler list
            """
        for (func, desc, args) in func_list:
            if desc == "doc":
                desc = func.__doc__
            c = Code(func, func.__doc__, args)
            print(func)
            self.append(Code(func))

    def add_external(self, func_list):
        """ Add a list of Code objects to the profiler list.

            - allows addition of outside functions to the profiler
            - uses function objects instead of code snippets
            """
        for code in func_list:
            print(code)
            self.append(Code(code))

    @timeit
    def test_func(self, **kwargs):
        """ Run a function for the purpose of timing the operation.

            - Only keyword arguments <kwargs> are passed.
            - Return value is the normal function retval.
            - Decorator function used for consistent timing and to collate results.
            """
        # remove kwargs used for testing
        func = kwargs.pop("func", None)
        reps = kwargs.pop("reps", 1)

        # repeat tests <reps> times
        for _ in range(reps):
            f = func(**kwargs)
        return f

    def profile(self, func_list, reps=10, **kwargs):
        for func in self:
            self.test_func(func=func, reps=reps, **kwargs)

    @staticmethod
    def arg_list(args):
        return ", ".join(f"{arg}={{{arg}}}" for arg in args)


def main():

    reps: int = 1000
    n: int = 10
    s: str = "test"

    func_list: List[tuple] = [
        (fibonacci, "doc", ["n"]),
        (fibonacci_2, "doc", ["n"]),
        (fibomemo, "doc", ["n"]),
        (fibo_deque, "doc", ["n"]),
        (fibo_fake_deque, "doc", ["n"]),
        (fibo_fake_deque_2, "doc", ["n"]),
    ]

    # TODO: replace fibonacci.__name__
    code = Profiler_.Code(fibonacci, fibonacci.__name__, ["n"])
    print(code)
    print(code.code)
    print(code.description)
    print(code.args)
    print(code.function())


if __name__ == "__main__":
    main()



import re
import string
from typing import List

RE_IDENTIFIER = r'[a-zA-Z_]?[a-zA-Z0-9_]*'

keywords: List[str] = [
    'and',
    'del',
    'from',
    'not',
    'while',
    'as',
    'elif',
    'global',
    'or',
    'with',
    'assert',
    'else',
    'if',
    'pass',
    'yield',
    'break'
    'except',
    'import',
    'print',
    'class',
    'exec',
    'in',
    'raise',
    'continue',
    'finally',
    'is',
    'return',
    'def',
    'for',
    'lambda',
    'try']


def identifier(s: str):
    ''' identifier ::=  (letter|"_") (letter | digit | "_")*
        letter     ::=  lowercase | uppercase
        lowercase  ::=  "a"..."z"
        uppercase  ::=  "A"..."Z"
        digit      ::=  "0"..."9"
        '''
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pathlib3 import Path
from dataclasses import dataclass, field, Field
from sys import path as python_path
import os

p = os.path(__file__)

print(p)


class PathPackError(IOError):
    ''' Error processing PathPack path. '''


class PathPack(Path):
    ''' pathlib.Path subclass to encapsulate more common methods.
    '''
    path_name: (Path) = '.'
    # _path: Field = field(init=False)

    def __i(self):
        Path().__init__()
        # self._path: Path = None
        pass
        # self._path: Path = self.path(self.path_name)

    @property
    def path(self):
        if not self._path:
            self._path = self.path(self.path_name)
        return self._path

    @path.setter
    def path(self, value):
        if isinstance(value, str):
            try:
                self._path = Path(value).resolve()
            except IOError:
                raise PathPackError(
                    f'Error converting {self.path_name} to Path.')
        if not self._path.exists():
            try:
                self._path().touch()
            except IOError:
                raise PathPackError(f'Error creating {self.path_name}')

    @property
    def path_str(self):
        return self.path.as_posix()

    def __str__(self):
        return self.path_str


if __name__ == '__main__':
    p = PathPack('.')

    script_path = Path(__file__).resolve().parents[0]
    here = Path().cwd()

    print(f"script location: {str(script_path[0]):<55.55}")
    print(f"current path:    {str(here):<55.55}")
    print(p)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """


class StrSubclass(str):
    def __getitem__(self, index):
        return StrSubclass(super().__getitem__(index))


class BytesSubclass(bytes):
    def __getitem__(self, index):
        return BytesSubclass(super().__getitem__(index))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" This is an example of ways to extract data (numerical, photos, files, etc)
    from online resources. Using the existing tools can be frustrating and
    leads to an enourmous burden of recordkeeping and 'fiddling' with settings
    on CLI apps. A more complete way to solve the 90% problem that leaves out
    the 10% of edge cases where stuff is strange leads to a strikingly simpler
    way of doing thing.

    Some highlights are:

    ## Efficient Resource gathering:

    - Enter a url or search terms(for Google search)
    - choose resources to gather (photos, videos, html, text, css, etc)
    - gather entire contents or just links?
    - decide where to put the output (Dropbox, Docker, Google, AWS, local)
    - choose output format (files with html directory, csv, json, pdf, etc)
    - a way to choose priority, schedule a time to perform, etc
    - a way to share progress and data with coworkers
    - click 'go' and let it work ... no micromanaging / no boilerplate
    - notify on completion (text, email, log, etc)

    ## Parts needed to accomplish these goals:

    - A way to log in to Google for Search, Drive, Calendar, Notifications, etc
    - A way to log all activities with timestamps to provide a record and trace
    - A data structure to store the request information
    - A data structure to store the request results
    - A way to log in to Dropbox, AWS, Azure, etc to create files and folders
    - A way to set local and remote permissions on the data collection
    - A way to attach dataset to events in calendars and share with collaborators
    - A way to log and notify upon completion of various steps and milestones.


    # example of a background photo that was difficult to extract:
    # https://images.unsplash.com/photo-1463620695885-8a91d87c53d0?ixlib=rb-1.2.1
    # original url and tag: <iframe src="https://workona.com/newtab/chrome?type=wka&amp;bg=img&amp;galleries=3587225%2C3587267%2C3587269%2C3587264%2C3587279%2C3588045%2C3587324&amp;color=%232B86F1&amp;sb=hvr&amp;clock=true&amp;militaryTime=false&amp;v=442071" frameborder="0" style="position: fixed; top: 0px; right: 0px; bottom: 0px; left: 0px; height: 100vh; width: 100vw;"></iframe>
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

""" #TODO -- ideas for maze solver

    - mazes with multiple solutions
    - find shortest
    - find lowest cost
    - make some color coded (or gradients) that label the 'cost' of traversing over a certain area
    - create a 3d maze
    - create an automated way of testing 'solver' algorithms

    """

if True:
    import numpy as np
    import pandas as PD
    from sys import stdout, stderr
    from typing import Dict, List


def indent(*args, **kwargs):
    print("  ", *args, **kwargs)


def eprint(*args, **kwargs):
    """ Print to sys.stderr for debugging. """
    print(*args, file=stderr)


def stats(x):
    """ Print common variable stats to <stderr> for debugging. """
    for i in (_ for _ in dir(x) if not _.startswith("_")):
        print("-" * 50)
        print(i)
        indent(type(dt))
        indent(eval("x." + str(i) + ""))
        indent(eval("type(x." + str(i) + ")"))
        print()


# class Location():
#     """ Location in a 3d space. """

#     def __init__(self, x: int = 0, y: int = 0, z: int = 0):
#         if x:
#             self.x: int = x
#         if y:
#             self.y: int = y
#         if z:
#             self.z: int = z


location = np.dtype(">i4")
dt = np.dtype(">i4")
stats(dt)


class Particle:
    """ #### Model of the properties of a location in 3d space. Some people
        may choose to think of it as a 'particle' of matter.

        data: unspecified type -- data for the properties of Particle

        properties: dict -- additional properties

        Base class for 3d physics models with physical properties and
        arbitrary granular precision.

        Examples of custom properties that could be added are:
        - density
        - temperature
        - 'squishyness'
        - velocity
        - charge
        - Energy
        - material (possibly a class with its own properties)
        """

    def __init__(self, data, properties: Dict, x=0, y=0, z=0):
        super().__init__(data, x=x, y=y, z=z)
        self.props = properties

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


class Mass:
    """ A collection of particles in 3d space.

        The properties are determined by the subclass of the <Location> class.
        """

    pass


class Area:
    """ A special subclass of <Mass> that ignores the z-axis by default.
    """

    pass


class Traveler:
    active: bool  # TODO or <done: bool> ???
    location: int


class Maze(Area):
    """ Maze solver

        Reference: based on Computerphile maze solver video.
        """

    start: (0, 5)  # this should be the top center of the maze

    crew: List[
        Traveler
    ] = []  # my crew of travelers who will go out and search the wastelands for a path to the water source

    def make_traveler(self, loc):
        self.crew.append(traveler(loc))

    def start_solving(self):
        make_traveler(START)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
# 'local imports'
from ._basic_colors import BasicColors
from ._css4 import CSS4_COLORS
from ._matplotlib import BASE_COLORS
from ._tableau_colors import TABLEAU_COLORS
from ._xkcd_colors import XKCD_COLORS

__all__ = [
    "BasicColors",
    "CSS4_COLORS",
    "BASE_COLORS",
    "TABLEAU_COLORS",
    "XKCD_COLORS",
]



"""
Small Sphinx extension intended to generate documentation for stub files.

It retrieves only the docstrings of "loguru/__init__.pyi", hence avoiding possible errors (caused by
missing imports or forward references). The stub file is loaded as a dummy module which contains
only the top-level docstring. All the formatting can therefore be handled by the "autodoc"
extension, which permits cross-reference.

The docstring of the stub file should list the available type hints and add short explanation of
their usage.

Warning: for some reason, the docs NEEDS to be re-generated for changes in the stub file to be taken
into account: ``make clean && make html``.
"""
import os
import sys
import types
import code


def get_module_docstring(filepath):
    with open(filepath) as file:
        source = file.read()

    co = compile(source, filepath, "exec")

    if co.co_consts and isinstance(co.co_consts[0], str):
        docstring = co.co_consts[0]
    else:
        docstring = None

    return docstring


def setup(app):
    module_name = "autodoc_stub_file.loguru"
    stub_path = os.path.join("..", "auto_loguru", "__init__.pyi")
    docstring = get_module_docstring(stub_path)
    module = types.ModuleType(module_name, docstring)
    sys.modules[module_name] = module



# 'Standard Library'
import re

from collections import namedtuple

# 'package imports'
from twitter.twitter_utils import enf_type

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


EndpointRateLimit = namedtuple(
    "EndpointRateLimit", ["limit", "remaining", "reset"]
)

ResourceEndpoint = namedtuple("ResourceEndpoint", ["regex", "resource"])


GEO_ID_PLACE_ID = ResourceEndpoint(
    re.compile(r"/geo/id/\d+"), "/geo/id/:place_id"
)
SAVED_SEARCHES_DESTROY_ID = ResourceEndpoint(
    re.compile(r"/saved_searches/destroy/\d+"), "/saved_searches/destroy/:id"
)
SAVED_SEARCHES_SHOW_ID = ResourceEndpoint(
    re.compile(r"/saved_searches/show/\d+"), "/saved_searches/show/:id"
)
STATUSES_RETWEETS_ID = ResourceEndpoint(
    re.compile(r"/statuses/retweets/\d+"), "/statuses/retweets/:id"
)
STATUSES_SHOW_ID = ResourceEndpoint(
    re.compile(r"/statuses/show"), "/statuses/show/:id"
)
USERS_SHOW_ID = ResourceEndpoint(re.compile(r"/users/show"), "/users/show/:id")
USERS_SUGGESTIONS_SLUG = ResourceEndpoint(
    re.compile(r"/users/suggestions/\w+$"), "/users/suggestions/:slug"
)
USERS_SUGGESTIONS_SLUG_MEMBERS = ResourceEndpoint(
    re.compile(r"/users/suggestions/.+/members"),
    "/users/suggestions/:slug/members",
)

NON_STANDARD_ENDPOINTS = [
    GEO_ID_PLACE_ID,
    SAVED_SEARCHES_DESTROY_ID,
    SAVED_SEARCHES_SHOW_ID,
    STATUSES_RETWEETS_ID,
    STATUSES_SHOW_ID,
    USERS_SHOW_ID,
    USERS_SUGGESTIONS_SLUG,
    USERS_SUGGESTIONS_SLUG_MEMBERS,
]


class RateLimit(object):

    """ Object to hold the rate limit status of various endpoints for
    the twitter.Api object.

    This object is generally attached to the API as Api.rate_limit, but is not
    created until the user makes a method call that uses _RequestUrl() or calls
    Api.InitializeRateLimit(), after which it get created and populated with
    rate limit data from Twitter.

    Calling Api.InitializeRateLimit() populates the object with all of the
    rate limits for the endpoints defined by Twitter; more info is available
    here:

        https://dev.twitter.com/rest/public/rate-limits

        https://dev.twitter.com/rest/public/rate-limiting

        https://dev.twitter.com/rest/reference/get/application/rate_limit_status

    Once a resource (i.e., an endpoint) has been requested, Twitter's response
    will contain the current rate limit status as part of the headers, i.e.::

        x-rate-limit-limit
        x-rate-limit-remaining
        x-rate-limit-reset

    ``limit`` is the generic limit for that endpoint, ``remaining`` is how many
    more times you can make a call to that endpoint, and ``reset`` is the time
    (in seconds since the epoch) until remaining resets to its default for that
    endpoint.

    Generally speaking, each endpoint has a 15-minute reset time and endpoints
    can either make 180 or 15 requests per window. According to Twitter, any
    endpoint not defined in the rate limit chart or the response from a GET
    request to ``application/rate_limit_status.json`` should be assumed to be
    15 requests per 15 minutes.

    """

    def __init__(self, **kwargs):
        """ Instantiates the RateLimitObject. Takes a json dict as
        kwargs and maps to the object's dictionary. So for something like:

        {"resources": {
                "help": {
                    /help/privacy": {
                        "limit": 15,
                        "remaining": 15,
                        "reset": 1452254278
                    }
                }
            }
        }

        the RateLimit object will have an attribute 'resources' from which you
        can perform a lookup like:

            api.rate_limit.get('help').get('/help/privacy')

        and a dictionary of limit, remaining, and reset will be returned.

        """
        self.__dict__["resources"] = {}
        self.__dict__.update(kwargs)

    @staticmethod
    def url_to_resource(url):
        """ Take a fully qualified URL and attempts to return the rate limit
        resource family corresponding to it. For example:

            >>> RateLimit.url_to_resource('https://api.twitter.com/1.1/statuses/lookup.json?id=317')
            >>> '/statuses/lookup'

        Args:
            url (str): URL to convert to a resource family.

        Returns:
            string: Resource family corresponding to the URL.
        """
        resource = urlparse(url).path.replace("/1.1", "").replace(".json", "")
        for non_std_endpoint in NON_STANDARD_ENDPOINTS:
            if re.match(non_std_endpoint.regex, resource):
                return non_std_endpoint.resource
        return resource

    def set_unknown_limit(self, url, limit, remaining, reset):
        return self.set_limit(url, limit, remaining, reset)

    def set_limit(self, url, limit, remaining, reset):
        """ If a resource family is unknown, add it to the object's
        dictionary. This is to deal with new endpoints being added to
        the API, but not necessarily to the information returned by
        ``/account/rate_limit_status.json`` endpoint.

        For example, if Twitter were to add an endpoint
        ``/puppies/lookup.json``, the RateLimit object would create a resource
        family ``puppies`` and add ``/puppies/lookup`` as the endpoint, along
        with whatever limit, remaining hits available, and reset time would be
        applicable to that resource+endpoint pair.

        Args:
            url (str):
                URL of the endpoint being fetched.
            limit (int):
                Max number of times a user or app can hit the endpoint
                before being rate limited.
            remaining (int):
                Number of times a user or app can access the endpoint
                before being rate limited.
            reset (int):
                Epoch time at which the rate limit window will reset.
        """
        endpoint = self.url_to_resource(url)
        resource_family = endpoint.split("/")[1]
        new_endpoint = {
            endpoint: {
                "limit": enf_type("limit", int, limit),
                "remaining": enf_type("remaining", int, remaining),
                "reset": enf_type("reset", int, reset),
            }
        }

        if not self.resources.get(resource_family, None):
            self.resources[resource_family] = {}

        self.__dict__["resources"][resource_family].update(new_endpoint)

        return self.get_limit(url)

    def get_limit(self, url):
        """ Gets a EndpointRateLimit object for the given url.

        Args:
            url (str, optional):
                URL of the endpoint for which to return the rate limit
                status.

        Returns:
            namedtuple: EndpointRateLimit object containing rate limit
            information.
        """
        endpoint = self.url_to_resource(url)
        resource_family = endpoint.split("/")[1]

        try:
            family_rates = self.resources.get(resource_family).get(endpoint)
        except AttributeError:
            return EndpointRateLimit(limit=15, remaining=15, reset=0)

        if not family_rates:
            self.set_unknown_limit(url, limit=15, remaining=15, reset=0)
            return EndpointRateLimit(limit=15, remaining=15, reset=0)

        return EndpointRateLimit(
            family_rates["limit"],
            family_rates["remaining"],
            family_rates["reset"],
        )



#!/usr/bin/env python

# Copyright 2007-2016 The Python-Twitter Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ------------------------------------------------------------------------
# Load the latest update for a Twitter user and output it as an HTML fragment
#

# 'future imports'
from __future__ import print_function

# 'Standard Library'
import argparse
import codecs
import sys

# 'package imports'
import twitter

from t import *

__author__ = "dewitt@google.com"

TEMPLATE = """
<div>
  <a href="http://twitter.com/{user}">Twitter</a>: {user}<br>
  {tweet_text}<br>
  <a href="http://twitter.com/{user}/statuses/{status_id}">Posted {tweet_created}</a>
</div>
"""


def main(**kwargs):
    api = twitter.Api(CONSUMER_KEY, CONSUMER_SECRET, ACCESS_KEY, ACCESS_SECRET)

    if kwargs["user_id"] is not None:
        statuses = api.GetUserTimeline(user_id=kwargs["user_id"])
    elif kwargs["screenname"] is not None:
        statuses = api.GetUserTimeline(screen_name=kwargs["screenname"])

    if kwargs["output_file"] is not None:
        with open(kwargs["output_file"], "w+") as f:
            for status in statuses:
                f.write(status)
    else:
        for status in statuses:
            print(
                TEMPLATE.format(
                    user=status.user.screen_name,
                    tweet_text=status.text,
                    status_id=status.id,
                    tweet_created=status.created_at,
                )
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--user-id", help="User id for which to return timeline"
    )
    group.add_argument(
        "--screenname", help="Screenname for which to return timeline"
    )
    parser.add_argument(
        "--output-file", help="Write to file instead of stdout"
    )
    args = parser.parse_args()
    if not (args.user_id or args.screenname):
        raise ValueError("You must specify one of user-id or screenname")
    if not all([CONSUMER_KEY, CONSUMER_SECRET, ACCESS_KEY, ACCESS_SECRET]):
        raise ValueError(
            "You must define CONSUMER_KEY, CONSUMER_SECRET, ACCESS_KEY, ACCESS_SECRET in t.py"
        )
    main(
        user_id=args.user_id,
        screenname=args.screenname,
        output_file=args.output_file,
    )



# 'package imports'
from autosys.text_utils.random_string import *

from typing import List, Sequence


def difference(a: Sequence, b: Sequence) -> (List):
    """ Return a list of items from `a` that are not in `b`.
    """
    return [item for item in set(a) if item not in set(b)]


a = [random_string() for i in range(100)]
b = [random_string() for i in range(100)]

# for i in dir():

# print(i)

print(dir())



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# from https://github.com/c4urself/bump2version

# 'Standard Library'
import errno
import logging
import os
import subprocess

from tempfile import NamedTemporaryFile

# 'package imports'
from autosys.exceptions import (
    MercurialDoesNotSupportSignedTagsException,
    WorkingDirectoryIsDirtyException)
from autosys.lister import *

logger = logging.getLogger(__name__)


class BaseVCS:

    _TEST_USABLE_COMMAND = None
    _COMMIT_COMMAND = None

    @classmethod
    def commit(cls, message, context, extra_args=None):
        extra_args = extra_args or []
        with NamedTemporaryFile("wb", delete=False) as f:
            f.write(message.encode("utf-8"))
        env = os.environ.copy()
        env["HGENCODING"] = "utf-8"
        for key in ("current_version", "new_version"):
            env[str("BUMPVERSION_" + key.upper())] = str(context[key])
        try:
            subprocess.check_output(
                cls._COMMIT_COMMAND + [f.name] + extra_args, env=env
            )
        except subprocess.CalledProcessError as exc:
            err_msg = "Failed to run {}: return code {}, output: {}".format(
                exc.cmd, exc.returncode, exc.output
            )
            logger.exception(err_msg)
            raise exc
        finally:
            os.unlink(f.name)

    @classmethod
    def is_usable(cls):
        try:
            return (
                subprocess.call(
                    cls._TEST_USABLE_COMMAND,
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                )
                == 0
            )
        except OSError as e:
            if e.errno in (errno.ENOENT, errno.EACCES):
                return False
            raise


class Git(BaseVCS):

    _TEST_USABLE_COMMAND = ["git", "rev-parse", "--git-dir"]
    _COMMIT_COMMAND = ["git", "commit", "-F"]
    _GIT_ADD_ALL = ["git", "add", "-all"]
    _GIT_COMMIT_AUTO = [
        "git",
        "commit",
        "-m",
        "The Great Gitspy autoformatting / refactoring.",
    ]
    _GIT_PUSH = ["git", "push", "--signed"]
    _GIT_PUSH_SUP = "git push --set-upstream origin $(git_current_branch)"

    @classmethod
    def assert_nondirty(cls):
        lines = [
            line.strip()
            for line in subprocess.check_output(
                ["git", "status", "--porcelain"]
            ).splitlines()
            if not line.strip().startswith(b"??")
        ]

        if lines:
            raise WorkingDirectoryIsDirtyException(
                "Git working directory is not clean:\n{}".format(
                    b"\n".join(lines).decode()
                )
            )

    @classmethod
    def latest_tag_info(cls):
        try:
            # git-describe doesn't update the git-index, so we do that
            subprocess.check_output(["git", "update-index", "--refresh"])

            # get info about the latest tag in git
            describe_out = (
                subprocess.check_output(
                    [
                        "git",
                        "describe",
                        "--dirty",
                        "--tags",
                        "--long",
                        "--abbrev=40",
                        "--match=v*",
                    ],
                    stderr=subprocess.STDOUT,
                )
                .decode()
                .split("-")
            )
        except subprocess.CalledProcessError:
            logger.debug("Error when running git describe")
            return {}

        info = {}

        if describe_out[-1].strip() == "dirty":
            info["dirty"] = True
            describe_out.pop()

        info["commit_sha"] = describe_out.pop().lstrip("g")
        info["distance_to_latest_tag"] = int(describe_out.pop())
        info["current_version"] = "-".join(describe_out).lstrip("v")

        return info

    @classmethod
    def add_path(cls, path):
        subprocess.check_output(["git", "add", "--update", path])

    @classmethod
    def tag(cls, sign, name, message):
        command = ["git", "tag", name]
        if sign:
            command += ["-s"]
        if message:
            command += ["--message", message]
        subprocess.check_output(command)


hr(n=60)
print(dir())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" # TODO -- @update `AutoSys` package
        copyright (c) 2018 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`


        """

# 'Standard Library'
from codecs import open as codec_open
from pathlib import Path

HERE = Path(__file__).resolve().parent


@log_it
def load_file_contents(file_path, as_list=True):
    """ Load file as string or list.

        Inspiration and portions of `load_file_contents`:
            :copyright: Copyright 2017 Andrew Pinkham
            :website: https://github.com/jambonrose/roman-numerals
            :license: Simplified BSD, see LICENSE_BSD for details.
        """
    # abs_file_path = path.join(HERE, file_path)
    try:
        abs_file_path = HERE / file_path
        d = c
    except:
        try:
            log.error(e)
        except:
            pass
        raise
    with codec_open(abs_file_path, encoding="utf-8") as fp:
        if as_list:
            return fp.read().splitlines()
        return fp.read()


print(load_file_contents("codec_open.py", as_list=False))
""" LICENSE_BSD

 BSD 2-Clause License

 Copyright (c) 2017, Andrew Pinkham
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" webpage - a class <WebPage> that represents a document and . It is designed for online documents that will be downloaded upon demand rather than stored in a database, but a database option is available for local implementation.

    Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
__license__ = "MIT"

# !------------------------------------- Imports
if True:  # standard library imports
    import re
    import sys

if True:  # standard library specifics
    from collections import deque
    from contextlib import closing
    from datetime import datetime
    from io import TextIOWrapper
    from sys import argv
    from time import perf_counter_ns as _timer
    from typing import (
        Any,
        Generator,
        List,
        Sequence,
        Dict,
        MutableSequence,
        Tuple,
    )

if True:  # outside requirements
    from requests import get
    from requests import Response
    from requests.exceptions import RequestException
    from bs4 import BeautifulSoup
    from bs4.element import ResultSet


DEFAULT_PARSER = "lxml"
DEFAULT_ENCODING = "utf-8"


class WebPageError(TypeError):
    """  # Exception raised for errors in the WebPage class. """

    pass


def is_good_response(resp):
    """
    Returns True if the response seems to be HTML, False otherwise.
    """
    content_type = resp.headers["Content-Type"].lower()
    return (
        resp.status_code == 200
        and content_type is not None
        and content_type.find("html") > -1
    )


def log_error(e):
    """
    It is always a good idea to log errors.
    This function just prints them, but you can
    make it do anything.
    """
    print(e)


def simple_get(url):
    """ Attempts to get the content at `url` by making an HTTP GET request.
        If the content-type of response is some kind of HTML/XML, return the
        text content, otherwise return None.
        """
    try:
        with closing(get(url, stream=True)) as resp:
            if is_good_response(resp):
                return resp.content
            else:
                return None

    except RequestException as e:
        log_error("Error during requests to {0} : {1}".format(url, str(e)))
        return None


class WebPage:
    """ A local representation of a webpage. This object is able to analyze
        itself, share data, and perform its duties unsupervised.

        parameters

        url: the source of the original page. Use the <refresh> command to get
        an updated version.
        """

    # >>> r = requests.get('https://api.github.com/user',
    import requests

    def __init__(self, url: str):
        self.url = url
        self.resp = requests.get(url)
        self.last_status: int = 0
        self._timestamp: datetime = datetime.now()
        # self.tags = {}
        # self.auth: Tuple[str, str] = ('user', 'pass')
        # auth = ('user', 'pass')
        super().__init__()

    def __iter__(self):
        return self.tags

    def __next__(self):
        try:
            yield (t for t in self.soup.find_all())
        except:
            raise StopIteration

        # if self.n <= self.max:
        # result = 2 ** self.n
        # self.n += 1
        #     return result
        # else:
        #     raise StopIteration

    # ------------------------------------------ properties
    @property
    def soup(self):
        # return BeautifulSoup()
        return BeautifulSoup(self.text, features=DEFAULT_PARSER)

    @property
    def tags(self) -> ResultSet:
        try:
            yield (t for t in self.soup.find_all())
        except:
            yield []

    @property
    def links(self):
        yield (t for t in self.soup.findAll("a", href=True))

    @property
    def images(self):
        yield (t for t in self.soup.findAll("img"))

    @property
    def text(self) -> str:
        return self.resp.text if self.resp.text else ""

    @property
    def status(self) -> int:
        return self.resp.status_code if self.resp.status_code else -1

    @property
    def timestamp(self) -> str:
        try:
            return datetime.fromtimestamp(self._timestamp).isoformat()
        except:
            return ""

    # ------------------------------------------ properties

    def get_url_content(self, url: str) -> str:
        """ Return decoded contents from <url> using default parameters. """
        self = requests.get(url)
        self.history.append()
        self.last_status = self.status_code
        if self.last_status == 200:
            return self.text
        else:
            return ""

    def tag_find(
        self,
        tag_name: str,
        attrs_pass: Dict,
        parser_pass: str = DEFAULT_PARSER,
    ) -> List:
        """ Find matching tags from url. """
        return self.soup.findAll(name=tag_name, attrs=attrs_pass)

    def tag_list(self, tag_name: str):
        if self.dirty:
            self.tags = Counter(self.text)
            return self.tags

    def to_markdown(self):
        pass

    def to_json(self):
        return json.dumps(self.text)

    def stats(self):
        pass

    def tags(
        self,
        name=None,
        attrs={},
        recursive=True,
        text=None,
        limit=None,
        **kwargs,
    ):
        return self.soup.findAll(
            name=name,
            attrs=attrs,
            recursive=recursive,
            text=text,
            limit=limit,
            **kwargs,
        )

    def __str__(self):
        return self.resp.url if self.resp.url else ""

    def __repr__(self):
        return f"response = {self.resp.status_code} for {len(self.resp.content)} bytes from {self.url}."

    def log_error(self, e: RequestException):
        """
            It is always a good idea to log errors.
            This function just prints them, but you can
            make it do anything.
            """
        print(e, file=sys.stderr)

    def is_valid(self, r: Response = None) -> bool:
        """ Return true if an HTTP response status is 200.

            The default response is <self.resp>.
            """
        if not r:
            r = self.resp
        return r.status_code == 200

    def is_html(self, r: Response = None) -> bool:
        """ Returns True if an HTTP response 'seems' to be HTML, False otherwise.

            The default response is <self>. Use parameter to test other urls.
            """
        if not r:
            r = self
        try:
            requests.Response.headers
            content_type = r.headers["Content-Type"].lower()
            return (
                r.status_code == 200
                and content_type is not None
                and content_type.find("html") > -1
            )
        except:
            return False

    # @time_it
    def _get(self, url: str = "", html: bool = True) -> (Response, Exception):
        """ Return the response an HTTP GET request to url.

            The default url is <self.url>, but outside url's can be tested as needed by setting <url>.

            If <html> is True, test the content-type of the response is some kind of HTML/XML and return the response, otherwise return None. (This is the default.)
            """
        # TODO check for dirty, refresh time, cookie, etc

        if not url:
            url = self.url
        try:
            with _closing(_get(url, stream=True)) as r:
                if self.is_valid(r):
                    if html:
                        if self.is_html(r):
                            return r
                    else:
                        return r
                else:
                    return None

        except RequestException as e:
            log_error(
                "Error during requests to {0} : {1}".format(self.url, str(e))
            )
            return None


class WebPageDeque(deque):
    """ Stores and maintains a set of web pages. """

    DEFAULT_WEBPAGESET_SIZE = 2000  # maximum number of pages
    # TODO this should be 'maximum size' ... not count
    # ... and class should check it's own size

    def __init__(
        self,
        iterable: MutableSequence,
        maxlen: int = 0,
        check_links: bool = True,
        image_storage: str = "",
    ):
        if maxlen < 1:
            maxlen = DEFAULT_WEBPAGESET_SIZE
        self.maxlen = maxlen
        self.check_links = check_links
        self.store = True if image_storage else False
        self.image_storage = image_storage
        super().__init__(iterable, maxlen)

    def count(self, needle=r"<a.*>"):
        """ count tags, emails, ... whatever from the entire set """
        pass

    def common(self):
        """ find items that this pageset has in common. """
        pass

    def append(self, x):
        """ Only append WebPage instances. """
        if isinstance(x, WebPage):
            super().append(x)
        else:
            if SET_DEBUG:
                db_print()
            else:
                raise (WebPageError)

    def size_check(self):
        print(self.__sizeof__())


sample_urls = [
    "http://foo.com/blah_blah",
    "http://foo.com/blah_blah/",
    "(Something like http://foo.com/blah_blah)",
    "http://foo.com/blah_blah_(wikipedia)",
    "(Something like http://foo.com/blah_blah_(wikipedia))",
    "http://foo.com/blah_blah.",
    "http://foo.com/blah_blah/.",
    "<http://foo.com/blah_blah>",
    "<http://foo.com/blah_blah/>",
    "http://foo.com/blah_blah,",
    "http://www.example.com/wpstyle/?p=364.",
    "http://✪df.ws/123",
    "rdar://1234",
    "rdar:/1234",
    "http://userid:password@example.com:8080",
    "http://userid@example.com",
    "http://userid@example.com:8080",
    "http://userid:password@example.com",
    "http://example.com:8080 x-yojimbo-item://6303E4C1-xxxx-45A6-AB9D-3A908F59AE0E",
    "message://%3c330e7f8409726r6a4ba78dkf1fd71420c1bf6ff@mail.gmail.com%3e",
    "http://➡.ws/䨹",
    "www.➡.ws/䨹",
    "<tag>http://example.com</tag>",
    "Just a www.example.com link.",
]

RE_STR_URL_MATCH = r'(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’]))'
RE_PATTERN_URL_MATCH = re.compile(RE_STR_URL_MATCH)


def re_matches(s, pat=RE_PATTERN_URL_MATCH):
    """ Return True if s matches pattern.
        Default is url pattern.
        """
    return re.match(pat, s)


def is_url(url) -> bool:
    """ regex url match
        https://daringfireball.net/2009/11/liberal_regex_for_matching_urls
        initial pattern:
        \b(([\w-]+://?|www[.])[^\s()<>]+(?:\([\w\d]+\)|([^[:punct:]\s]|/)))
        superseded to:
        https://daringfireball.net/2010/07/improved_regex_for_matching_urls
        """
    return re_matches(url, RE_PATTERN_URL_MATCH)


def _test_(url):
    w = WebPage(url)
    r = w.status
    print(w)
    print(r)
    print(w.is_html(r))
    print()

    print(w)
    # content_type = w.headers['Content-Type'].lower()
    # print(content_type)
    # print(w.resp.headers['Content-Type'].lower())


def _main_():
    """
    CLI script main entry point.
    """
    url = "https://realpython.com/blog/"

    _test_(url)

    print(simple_get(url=url))

    w = WebPage("https://realpython.com/blog/")

    print(w.text)

    url = "https://realpython.com/blog/nope-not-gonna-find-it"

    print(WebPage(url))


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_()



# Ref: https://wiki.python.org/moin/PythonSpeed/PerformanceTips

# Only if you are using older versions of Python (before 2.4) does the following advice from Guido van Rossum apply:

# An alternative way to speed up sorts is to construct a list of tuples whose first element is a sort key that will sort properly using the default comparison, and whose second element is the original list element. This is the so-called Schwartzian Transform, also known as DecorateSortUndecorate(DSU).

# Suppose, for example, you have a list of tuples that you want to sort by the n-th field of each tuple. The following function will do that.


def sortby(somelist, n):
    nlist = [(x[n], x) for x in somelist]
    nlist.sort()
    return [val for (key, val) in nlist]


# Matching the behavior of the current list sort method (sorting in place) is easily achieved as well:


def sortby_inplace(somelist, n):
    somelist[:] = [(x[n], x) for x in somelist]
    somelist.sort()
    somelist[:] = [val for (key, val) in somelist]
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" py_code.py - print the number of lines
        of python code in a directory. """
# 'Standard Library'
import sys

from pathlib import Path

# 'package imports'
import autosys.as_ansi as ansi

from typing import Dict, List

ce = ansi.color_encode
cp = ansi.color_print

FG = ansi.FG_DICT


ATTN = FG["ATTN"]
RESET = FG["RESET"]
FORMAT_STR: str = f"Lines of code matching glob pattern '{ATTN}{{:<6}}{RESET}':{ATTN}{{:>10}}{RESET}"


# def c_print(c: str, args):
#     try:
#         print(FG[c], args, FG["RESET"])
#     except:
#         print(args)
#         print(FG["RESET"])


def get_suffix_list():
    pass


def file_len(file_name: str, encoding: str = "utf8") -> int:
    i: int = -1
    try:
        with open(file_name, encoding=encoding) as f:
            i = -1
            for i, line in enumerate(f):
                pass
    except (UnicodeDecodeError, FileNotFoundError):
        pass
    return i + 1


def code_lines(pattern: str, recursive: bool = True) -> int:
    if not recursive:
        return sum(file_len(f) for f in p.glob(pattern) if not f.is_dir())
    else:
        return sum(file_len(f) for f in p.rglob(pattern) if not f.is_dir())


def print_code(pattern: str):
    print(FORMAT_STR.format(pattern, code_lines(pattern)))


if __name__ == "__main__":
    if len(sys.argv) > 1:
        pattern = sys.argv[1]
    else:
        pattern_list = [
            "py",
            "c",
            "sh",
            "js",
            "bas",
            "pas",
            "php",
            "css",
            "html",
            "cpp",
            "h",
            "md",
            "rst",
        ]
    p = Path().cwd()
    p = p.resolve()  # similar to os.path.abspath()
    cp("MAIN", "-~*#^#*~-" * 8)
    cp("MAIN", "Current Path: {}".format(p))
    cp("MAIN", "-~*#^#*~-" * 8)
    pattern_list = [
        "py",
        "c",
        "sh",
        "js",
        "bas",
        "pas",
        "php",
        "css",
        "html",
        "cpp",
        "h",
        "md",
        "rst",
    ]
    for pattern in pattern_list:
        print_code("*." + pattern)


"""
References:

original: https://towardsdatascience.com/bite-sized-python-recipes-52cde45f1489

"""
